# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import GObject  # @UnresolvedImport

from ....observers import ListObserver, ListItemObserver

from .base_models import BaseObjectListStore
from weakref import WeakKeyDictionary

class ObjectListStore(BaseObjectListStore):
    """
        GenericTreeModel implementation that wraps a python list of 
        mvc model objects. In addition, it expects all objects 
        to be of a certain type, which needs to be passed to the __init__ as 
        the first argument.This way, the wrapper can inspect the type and
        find out what properties can be represented as columns and report this
        to Gtk.
    """

    # PROPERTIES:
    _deleted_paths = None

    @property
    def _data(self):
        if self._model is not None:
            return getattr(self._model, self._prop_name, None)
        else:
            return []

    def is_wrapping(self, model, prop_name):
        return self._model == model and self._prop_name == prop_name

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, model, prop):
        BaseObjectListStore.__init__(self, prop.data_type)
        self._model = model
        self._prop_name = prop.label

        self._deleted_paths = []

        self._observer = ListObserver(
            self.on_item_inserted, self.on_item_deleted,
            on_deleted_before=self.on_item_deleted_before,
            prop_name=self._prop_name, model=self._model
        )

        self._list_item_observers = WeakKeyDictionary()
        for item in self._data:
            self._observe_item(item)

    def _observe_item(self, item):
        obs = ListItemObserver(self.on_item_changed, model=item)
        self._list_item_observers[item] = obs

    def _unobserve_item(self, item):
        observer = self._list_item_observers.get(item, None)
        if observer is not None: observer.clear()

    def on_item_changed(self, item):
        itr = self.create_tree_iter(item)
        path = self.get_path(itr)
        try:
            self.row_changed(path, itr)
        except TypeError as err:
            err.args += ("when emitting row_changed using:", path, itr)
            raise

    def on_item_inserted(self, item):
        try:
            itr = self.create_tree_iter(item)
            path = self.get_path(itr)
            self._observe_item(item)
            self.row_inserted(path, itr)
        except ValueError:
            logger.debug("Invalid rowref passed: %s", item)
            pass # invalid rowref

    def on_item_deleted_before(self, item):
        self._unobserve_item(item)
        self._deleted_paths.append((self._data.index(item),))

    def on_item_deleted(self, item):
        for path in self._deleted_paths:
            self.row_deleted(path)
        self._deleted_paths = []


    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def on_get_iter(self, path):
        try:
            return self._data[path[0]]
        except IndexError:
            return None

    def on_get_path(self, rowref):
        try:
            return (self._data.index(rowref),)
        except ValueError:
            logger.exception("ValueError in on_get_path of %s caused by %s" % (self, rowref))

    def set_value(self, itr, column, value):
        user_data = self.get_user_data(itr)
        setattr(user_data, self._columns[column][0], value)
        self.row_changed(self.get_path(itr), itr)

    def on_get_value(self, rowref, column):
        value = getattr(rowref, self._columns[column][0])
        try:
            default = self._columns[column][1]()
        except TypeError:
            default = ""
        return value if value is not None else default

    def on_iter_next(self, rowref):
        n, = self.on_get_path(rowref)
        try:
            return self._data[n + 1]
        except IndexError:
            pass

    def on_iter_children(self, rowref):
        if rowref:
            return None
        if self._data:
            return self.on_get_iter((0,))
        return None

    def on_iter_has_child(self, rowref):
        if rowref:
            return False
        if len(self._data) > 0:
            return True
        return False

    def on_iter_n_children(self, rowref):
        if rowref:
            return 0
        return len(self._data)

    def on_iter_nth_child(self, parent, n):
        if parent:
            return None
        if n < 0 or n >= len(self._data):
            return None
        return self._data[n]

    def on_iter_parent(self, rowref):
        return None

    pass # end of class

GObject.type_register(ObjectListStore) # @UndefinedVariable



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, string

import numpy as np
import xml.etree.ElementTree as ET

from pyxrd.mixture.models import Mixture
from pyxrd.project.models import Project
from pyxrd.specimen.models import Specimen
from pyxrd.phases.models import Phase
from pyxrd.atoms.models import Atom

def safe_float(num):
    return float(num.replace(",", "."))

def create_project_from_sybilla_xml(filename, **kwargs):
    """
        Creates a new project structure from a Sybilla XML file.
        Some information (e.g. the actual XRD pattern) is not present and will
        still need to be imported manually.
    """

    tree = ET.parse(filename)
    root = tree.getroot()
    basename = os.path.basename(filename)

    # Create the project:
    if "name" in kwargs: kwargs.pop("name")
    if "layout_mode" in kwargs: kwargs.pop("layout_mode")
    project = Project(name=basename, layout_mode="FULL", **kwargs)

    # Add a specimen:
    specimen = Specimen(name=basename, parent=project)
    project.specimens.append(specimen)

    # Add a mixture:
    mixture = Mixture(name=basename, auto_run=False, parent=project)
    mixture.add_specimen_slot(specimen, 1.0, 0.0)
    project.mixtures.append(mixture)

    with project.data_changed.ignore():
        with mixture.data_changed.ignore():

            for child in root:
                if child.tag == "basic_params":
                    # Goniometer parameters:
                    step_size = safe_float(child.attrib['step_size'])
                    wavelength = safe_float(child.attrib['lambda']) / 10.0
                    steps = int(1 + (specimen.goniometer.max_2theta - specimen.goniometer.min_2theta) / step_size)

                    specimen.goniometer.min_2theta = safe_float(child.attrib['min2theta'])
                    specimen.goniometer.max_2theta = safe_float(child.attrib['max2theta'])
                    specimen.goniometer.steps = steps
                    specimen.goniometer.wavelength = wavelength
                elif child.tag == "diffractometer":
                    # Some more goniometer parameters, and specimen parameters:
                    specimen.goniometer.radius = safe_float(child.attrib['gonio_radius'])
                    specimen.goniometer.divergence = safe_float(child.attrib['diverg_slit'])
                    specimen.goniometer.soller1 = safe_float(child.attrib['Soller1'])
                    specimen.goniometer.soller2 = safe_float(child.attrib['Soller2'])
                    specimen.sample_length = safe_float(child.attrib['sample_length'])
                elif child.tag == "content":
                    # Content tag contains 'Mixture' data
                    for xmlPhaseContent in child:
                        name = xmlPhaseContent.attrib['name']
                        fraction = safe_float(xmlPhaseContent.attrib['content']) / 100.
                        mixture.add_phase_slot(name, fraction)
                elif child.tag == "mixture":
                    # Mixture tag corresponds with the phases in the project level,
                    # not an actual Mixture object:
                    for xmlPhase in child:
                        name = xmlPhase.attrib['name']
                        sigma = xmlPhase.attrib['sigma_star']
                        csds = safe_float(xmlPhase.find('distribution').attrib['Tmean'])
                        G = 1
                        R = 0
                        W = [1.0, ]
                        if xmlPhase.attrib['type'] != 'mono':
                            prob = xmlPhase.find('probability')
                            G = int(prob.attrib['no_of_comp'])
                            R = int(prob.attrib['R'])

                        # create phase and add to project:
                        phase = Phase(name=name, sigma_star=sigma, G=G, R=R, parent=project)
                        phase.CSDS_distribution.average = csds
                        project.phases.append(phase)

                        # set probability:
                        if R == 0 and G != 1:
                            xmlW = prob.find('W')
                            W = np.array([ float(int(safe_float(xmlW.attrib[string.ascii_lowercase[i]]) * 1000.)) / 1000. for i in range(G) ])
                            for i in range(G - 1):
                                setattr(phase.probabilities, "F%d" % (i + 1), W[i] / np.sum(W[i:]))
                        if R == 1 and G == 2:
                            pass  # TODO
                        # ... TODO other probs

                        # parse components:
                        for i, layer in enumerate(xmlPhase.findall("./layer_and_edge/layer")):

                            component = phase.components[i]
                            component.name = layer.attrib['name']

                            component.d001 = safe_float(layer.attrib['d_spacing']) / 10.0
                            component.default_c = safe_float(layer.attrib['d_spacing']) / 10.0
                            component.delta_c = safe_float(layer.attrib['d_spacing_delta']) / 10.0

                            component.ucp_b.value = 0.9

                            component.ucp_a.factor = 0.57735
                            component.ucp_a.prop = (component, 'cell_b')
                            component.ucp_a.enabled = True


                            atom_type_map = {
                                # "NH4": "FIXME"
                                "K": "K1+",
                                "O": "O1-",
                                "Si": "Si2+",
                                "OH": "OH1-",
                                "Fe": "Fe1.5+",
                                "Al": "Al1.5+",
                                "Mg": "Mg1+",
                                "H2O": "H2O",
                                "Gly": "Glycol",
                                "Ca": "Ca2+",
                                "Na": "Na1+",
                            }

                            # add atoms:
                            fe_atom = None
                            encountered_oxygen = False
                            for atom in layer.findall("atom"):
                                atom_type_name = atom_type_map.get(atom.attrib['type'], None)
                                if atom_type_name:
                                    if atom_type_name == "O1-":
                                        # From this point we're dealing with layer atoms
                                        encountered_oxygen = True
                                    atom = Atom(
                                        name=atom.attrib['type'],
                                        default_z=safe_float(atom.attrib['position']) / 10.0,
                                        pn=safe_float(atom.attrib['content']),
                                        atom_type_name=atom_type_name,
                                        parent=component
                                    )
                                    if encountered_oxygen:
                                        component.layer_atoms.append(atom)
                                    else:
                                        component.interlayer_atoms.append(atom)
                                    atom.resolve_json_references()
                                    # Assume this is the octahedral iron...
                                    if encountered_oxygen and atom_type_name == "Fe1.5+":
                                        fe_atom = atom

                            # Set the atom relation
                            if fe_atom is not None:
                                component.ucp_b.constant = 0.9
                                component.ucp_b.factor = 0.0043
                                component.ucp_b.prop = (fe_atom, 'pn')
                                component.ucp_b.enabled = True

                pass # end of if
            pass # end of for


            # Map phases onto mixture names:
            for phase in project.phases:
                for slot, phase_name in enumerate(mixture.phases):
                    if phase.name == phase_name:
                        mixture.set_phase(0, slot, phase)

    return project



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test, mock_settings

from pyxrd.phases.models import Phase
from pyxrd.specimen.models import Specimen
from pyxrd.project.models import Project
from pyxrd.mixture.models import Mixture
from _collections import defaultdict

__all__ = [
    'TestMixture',
]

# Requires properly working:
#  - Phase
#  - Specimen
#  - Project

class TestMixture(unittest.TestCase):

    atom_type = None

    def setUp(self):
        mock_settings()
        self.project = Project(name="TestProject")
        self.mixture = Mixture(name="TestMixture", parent=self.project)

    def tearDown(self):
        del self.mixture
        del self.project

    def test_not_none(self):
        self.assertIsNotNone(self.mixture)

    def test_data_object(self):
        self.assertIsNotNone(self.mixture.data_object)

    test_name = create_object_attribute_test("mixture", "name", "Test Name")

    def test_parent(self):
        parent_project = Project(name="Parent2")
        self.mixture.parent = parent_project
        self.assertEqual(self.mixture.parent, parent_project)

    def test_add_phase_slot(self):
        index = self.mixture.add_phase_slot("TestPhase", 0.5)
        self.assertEqual(index, 0, "Adding a phase slot should return the correct index!")

    def test_add_specimen_slot(self):
        index = self.mixture.add_specimen_slot(None, 1.0, 0)
        self.assertEqual(index, 0, "Adding a specimen slot should return the correct index!")

    def test_add_order1(self):
        """Test if addition works when 1st specimen slot is added before 1st phase slot"""
        self.mixture.add_specimen_slot(None, 1.0, 0)
        self.assertEqual(len(self.mixture.specimens), 1)
        self.assertEqual(len(self.mixture.phases), 0)
        self.assertEqual(len(self.mixture.fractions), 0)
        self.assertEqual(self.mixture.phase_matrix.shape, (1, 0))
        self.mixture.add_phase_slot("TestPhase", 0.5)
        self.assertEqual(len(self.mixture.specimens), 1)
        self.assertEqual(len(self.mixture.phases), 1)
        self.assertEqual(len(self.mixture.fractions), 1)
        self.assertEqual(self.mixture.phase_matrix.shape, (1, 1))
        self.assertEqual(self.mixture.phase_matrix[0, 0], None)

    def test_add_order2(self):
        """Test if addition works when 1st phase slot is added before 1st specimen slot"""
        self.mixture.add_phase_slot("TestPhase", 0.5)
        self.assertEqual(len(self.mixture.specimens), 0)
        self.assertEqual(len(self.mixture.phases), 1)
        self.assertEqual(len(self.mixture.fractions), 1)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 1))
        self.mixture.add_specimen_slot(None, 1.0, 0)
        self.assertEqual(len(self.mixture.specimens), 1)
        self.assertEqual(len(self.mixture.phases), 1)
        self.assertEqual(len(self.mixture.fractions), 1)
        self.assertEqual(self.mixture.phase_matrix.shape, (1, 1))
        self.assertEqual(self.mixture.phase_matrix[0, 0], None)

    def test_add_multiple(self):
        """Test if addition for multiple phases and specimens works as expected"""
        self.mixture.add_phase_slot("TestPhase", 0.5)
        self.mixture.add_specimen_slot(None, 1.0, 0)
        self.mixture.add_specimen_slot(None, 1.0, 0)
        self.mixture.add_phase_slot("TestPhase2", 0.5)
        self.mixture.add_phase_slot("TestPhase3", 0.5)
        self.assertEqual(len(self.mixture.specimens), 2)
        self.assertEqual(len(self.mixture.phases), 3)
        self.assertEqual(len(self.mixture.fractions), 3)
        self.assertEqual(self.mixture.phase_matrix.shape, (2, 3))

    def test_del_phase_slot(self):
        """Test if deleting a phase works as expected"""
        self.mixture.add_phase_slot("TestPhase1", 0.1)
        self.mixture.add_phase_slot("TestPhase2", 0.1)
        self.mixture.del_phase_slot(1)
        self.assertEqual(len(self.mixture.phases), 1)
        self.assertEqual(len(self.mixture.fractions), 1)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 1))
        self.mixture.del_phase_slot(0)
        self.assertEqual(len(self.mixture.phases), 0)
        self.assertEqual(len(self.mixture.fractions), 0)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 0))

    def test_del_specimen_slot(self):
        """Test if deleting a specimen works as expected"""
        self.mixture.add_specimen_slot(None, 0.5, 0)
        self.mixture.add_specimen_slot(None, 0.5, 0)
        self.mixture.del_specimen_slot(1)
        self.assertEqual(len(self.mixture.specimens), 1)
        self.assertEqual(self.mixture.phase_matrix.shape, (1, 0))
        self.mixture.del_specimen_slot(0)
        self.assertEqual(len(self.mixture.specimens), 0)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 0))

    def test_del_phase_slot_by_name(self):
        self.mixture.add_phase_slot("TestPhase1", 0.1)
        self.mixture.del_phase_slot_by_name("TestPhase1")
        self.assertEqual(len(self.mixture.phases), 0)
        self.assertEqual(len(self.mixture.fractions), 0)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 0))

    def test_del_specimen_slot_by_object(self):
        dummy = Specimen(name="Test Specimen", parent=self.project)
        self.project.specimens.append(dummy)
        self.mixture.add_specimen_slot(dummy, 0.5, 0)
        self.mixture.del_specimen_slot_by_object(dummy)
        self.assertEqual(len(self.mixture.specimens), 0)
        self.assertEqual(self.mixture.phase_matrix.shape, (0, 0))

    def test_set_specimen(self):
        dummy = Specimen(name="Test Specimen", parent=self.project)
        self.project.specimens.append(dummy)
        self.mixture.add_specimen_slot(None, 0.5, 0)
        self.mixture.set_specimen(0, dummy)
        self.assertEqual(self.mixture.specimens[0], dummy)

    def test_unset_specimen(self):
        dummy = Specimen(name="Test Specimen", parent=self.project)
        self.project.specimens.append(dummy)
        self.mixture.add_specimen_slot(dummy, 0.5, 0)
        self.mixture.unset_specimen(dummy)
        self.assertEqual(self.mixture.specimens[0], None)

    def test_unset_phase(self):
        specimen = Specimen(name="Test Specimen", parent=self.project)
        self.project.specimens.append(specimen)
        self.mixture.add_specimen_slot(specimen, 0.5, 0)
        self.mixture.add_phase_slot("Test Phase1", 0.5)

        dummy = Phase(name="Test Phase", parent=self.project)
        self.project.phases.append(dummy)
        self.mixture.set_phase(0, 0, dummy)
        self.mixture.unset_phase(dummy)
        self.assertEqual(self.mixture.phase_matrix[0, 0], None)

    def test_randomize_empty_mixture(self):
        self.mixture.refinement.randomize()

    def _refinement_setup(self):
        # TODO maybe add some more variation in the type of Phases?
        specimen = Specimen(name="Test Specimen", parent=self.project)
        self.project.specimens.append(specimen)
        phase1 = Phase(name="Test Phase1", parent=self.project)
        self.project.phases.append(phase1)
        phase2 = Phase(name="Test Phase2", parent=self.project)
        self.project.phases.append(phase2)
        self.mixture.add_specimen_slot(specimen, 0.5, 0)
        self.mixture.add_phase_slot("Test Phase1", 0.5)
        self.mixture.add_phase_slot("Test Phase2", 0.5)
        self.mixture.set_phase(0, 0, phase1)
        self.mixture.set_phase(0, 1, phase2)

    def test_randomize(self):
        self._refinement_setup()

        # Mark the attribute(s) for refinement & get their values:
        refinables = []
        for node in self.mixture.refinables.iter_children():
            ref_prop = node.object
            if ref_prop.refinable:
                ref_prop.refine = True
                refinables.append((ref_prop, ref_prop.value))

        # Randomize:
        self.mixture.refinement.randomize()

        # Check all of them have been randomized:
        # It is possible (but unlikely) that the randomized value
        # is the same as the pre-randomized value. If so run this test again
        # to make sure it is really failing.
        for ref_prop, pre_val in refinables:
            self.assertNotEqual(pre_val, ref_prop.value)

    def test_auto_restrict_empy_mixture(self):
        self.mixture.refinement.auto_restrict()

    def test_auto_restrict(self):
        self._refinement_setup()

        # Mark the attribute(s) for refinement & get their values:
        refinables = []
        for node in self.mixture.refinables.iter_children():
            ref_prop = node.object
            if ref_prop.refinable:
                ref_prop.refine = True
                refinables.append((ref_prop, ref_prop.value))

        # Randomize:
        self.mixture.refinement.auto_restrict()

        # Check all of them have been restricted:
        for ref_prop, pre_val in refinables:
            self.assertEqual(pre_val * 0.8, ref_prop.value_min)
            self.assertEqual(pre_val * 1.2, ref_prop.value_max)


    # TODO:
    #  - set_data_object
    #  - optimize
    #  - apply_current_data_object
    #  - update
    #  - get_refinement_method
    #  - setup_refine_options

    pass # end of class

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os
import queue
import logging
from logging.handlers import QueueHandler, QueueListener

def setup_logging(basic=False, prefix=None):
    """
        Setup logging module.
    """
    from pyxrd.data import settings

    # Whether PyXRD should spew out debug messages
    debug = settings.DEBUG
    # Filename used for storing the logged messages
    log_file = settings.LOG_FILENAME
    # Flag indicating if a full logger should be setup (False) or
    # if simple, sparse logging is enough (True)
    basic = not settings.GUI_MODE

    fmt = '%(name)s - %(levelname)s: %(message)s'
    if prefix is not None:
        fmt = prefix + " " + fmt 

    if log_file is not None and not os.path.exists(os.path.dirname(log_file)):
        os.makedirs(os.path.dirname(log_file))

    if not basic:
        # Setup file log:
        file_handler = logging.FileHandler(log_file, 'w')
        disk_fmt = logging.Formatter(
            '%(asctime)s %(levelname)-8s %(name)-40s %(message)s',
            datefmt='%m-%d %H:%M')
        file_handler.setFormatter(disk_fmt)

        # Setup console log:
        log_handler = logging.StreamHandler()
        full = logging.Formatter(fmt)
        log_handler.setFormatter(full)

        # Setup queue handler:
        log_que = queue.Queue(-1)
        queue_handler = QueueHandler(log_que)
        queue_listener = QueueListener(log_que, file_handler, log_handler, respect_handler_level=True)
        queue_listener.start()
        
        # Add queue handler:
        logger = logging.getLogger('')
        logger.setLevel(logging.DEBUG if debug else logging.INFO)
        logger.addHandler(queue_handler)
    else:
        # Very basic output for the root object:
        logging.basicConfig(format=fmt)
        logger = logging.getLogger('')
        logger.addHandler(queue_handler)
        
    settings.FINALIZERS.append(queue_listener.stop)



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

from test.utils import create_object_attribute_test

from pyxrd.data import settings

from pyxrd.phases.models import Phase
from pyxrd.file_parsers.phase_parsers import JSONPhaseParser

__all__ = [
    'TestPhase',
]

class TestPhase(unittest.TestCase):

    phase = None

    def setUp(self):
        settings.initialize()
        self.phase = Phase(R=0, G=1)

    def tearDown(self):
        del self.phase

    def test_not_none(self):
        self.assertIsNotNone(self.phase)

    def test_data_object(self):
        self.assertIsNotNone(self.phase.data_object)

    def test_R_G(self):
        self.assertIsNotNone(Phase(R=0, G=1))
        self.assertIsNotNone(Phase(R=0, G=2))
        self.assertIsNotNone(Phase(R=0, G=3))
        self.assertIsNotNone(Phase(R=0, G=4))
        self.assertIsNotNone(Phase(R=0, G=5))
        self.assertIsNotNone(Phase(R=0, G=6))

        self.assertIsNotNone(Phase(R=1, G=2))
        self.assertIsNotNone(Phase(R=1, G=3))
        self.assertIsNotNone(Phase(R=1, G=4))

        self.assertIsNotNone(Phase(R=2, G=2))
        self.assertIsNotNone(Phase(R=2, G=3))

        self.assertIsNotNone(Phase(R=3, G=2))

    test_name = create_object_attribute_test("phase", "name", "Test Name")
    test_display_color = create_object_attribute_test("phase", "display_color", "#FF00FF")
    test_default_c = create_object_attribute_test("phase", "default_c", 0.646)
    test_sigma_star = create_object_attribute_test("phase", "sigma_star", 12.5)
    test_inherit_display_color = create_object_attribute_test("phase", "inherit_display_color", True)
    test_inherit_CSDS_distribution = create_object_attribute_test("phase", "inherit_CSDS_distribution", True)
    test_inherit_sigma_star = create_object_attribute_test("phase", "inherit_sigma_star", True)
    test_inherit_probabilities = create_object_attribute_test("phase", "inherit_probabilities", True)

    def test_import_export(self):
        from io import BytesIO
        phases = [Phase(R=0, G=1), Phase(R=1, G=2)]
        fn = BytesIO()
        Phase.save_phases(phases, filename=fn)
        loaded_phases = list(JSONPhaseParser.parse(fn))

        def strip_uuid(data):
            new_data = []
            for line in data.split('\n'):
                if "uuid" not in line:
                    new_data.append(line)
            return "\n".join(new_data)

        outp1 = [strip_uuid(phase.dump_object()) for phase in phases]
        outp2 = [strip_uuid(phase.dump_object()) for phase in loaded_phases]
        self.assertEqual(outp1, outp2)





    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import zipfile
from warnings import warn

from mvc.models.properties import (
    FloatProperty, SignalMixin, BoolProperty, StringProperty,
    LabeledProperty, SignalProperty, ObserveMixin, ListProperty,

)
from mvc.observers.list_observer import ListObserver

from pyxrd.generic.io import storables, Storable, COMPRESSION
from pyxrd.generic.models import DataModel
from pyxrd.generic.models.properties import InheritableMixin

from pyxrd.calculations.components import get_factors
from pyxrd.calculations.data_objects import ComponentData

from pyxrd.refinement.refinables.mixins import RefinementGroup
from pyxrd.refinement.refinables.metaclasses import PyXRDRefinableMeta
from pyxrd.refinement.refinables.properties import RefinableMixin

from pyxrd.atoms.models import Atom

from pyxrd.file_parsers.json_parser import JSONParser

from .atom_relations import AtomRelation
from .unit_cell_prop import UnitCellProperty

@storables.register()
class Component(RefinementGroup, DataModel, Storable, metaclass=PyXRDRefinableMeta):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Component"

    _data_object = None
    @property
    def data_object(self):
        weight = 0.0

        self._data_object.layer_atoms = [None] * len(self.layer_atoms)
        for i, atom in enumerate(self.layer_atoms):
            self._data_object.layer_atoms[i] = atom.data_object
            weight += atom.weight

        self._data_object.interlayer_atoms = [None] * len(self.interlayer_atoms)
        for i, atom in enumerate(self.interlayer_atoms):
            self._data_object.interlayer_atoms[i] = atom.data_object
            weight += atom.weight

        self._data_object.volume = self.get_volume()
        self._data_object.weight = weight
        self._data_object.d001 = self.d001
        self._data_object.default_c = self.default_c
        self._data_object.delta_c = self.delta_c
        self._data_object.lattice_d = self.lattice_d

        return self._data_object

    phase = property(DataModel.parent.fget, DataModel.parent.fset)

    # SIGNALS:
    atoms_changed = SignalProperty()

    # UNIT CELL DIMENSION SHORTCUTS:
    @property
    def cell_a(self):
        return self._ucp_a.value
    @property
    def cell_b(self):
        return self._ucp_b.value
    @property
    def cell_c(self):
        return self.d001

    # PROPERTIES:

    #: The name of the Component
    name = StringProperty(
        default="", text="Name",
        visible=True, persistent=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit the UCP a from :attr:`~linked_with`
    @BoolProperty(
        default=False, text="Inh. cell length a",
        visible=True, persistent=True, tabular=True,
    )
    def inherit_ucp_a(self):
        return self._ucp_a.inherited
    @inherit_ucp_a.setter
    def inherit_ucp_a(self, value):
        self._ucp_a.inherited = value

    #: Flag indicating whether to inherit the UCP b from :attr:`~linked_with`
    @BoolProperty(
        default=False, text="Inh. cell length b",
        visible=True, persistent=True, tabular=True,
    )
    def inherit_ucp_b(self):
        return self._ucp_b.inherited
    @inherit_ucp_b.setter
    def inherit_ucp_b(self, value):
        self._ucp_b.inherited = value

    #: Flag indicating whether to inherit d001 from :attr:`~linked_with`
    inherit_d001 = BoolProperty(
        default=False, text="Inh. cell length c",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit default_c from :attr:`~linked_with`
    inherit_default_c = BoolProperty(
        default=False, text="Inh. default length c",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit delta_c from :attr:`~linked_with`
    inherit_delta_c = BoolProperty(
        default=False, text="Inh. c length dev.",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit layer_atoms from :attr:`~linked_with`
    inherit_layer_atoms = BoolProperty(
        default=False, text="Inh. layer atoms",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit interlayer_atoms from :attr:`~linked_with`
    inherit_interlayer_atoms = BoolProperty(
        default=False, text="Inh. interlayer atoms",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether to inherit atom_relations from :attr:`~linked_with`
    inherit_atom_relations = BoolProperty(
        default=False, text="Inh. atom relations",
        visible=True, persistent=True, tabular=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )

    _linked_with_index = None
    _linked_with_uuid = None

    #: The :class:`~Component` this component is linked with
    linked_with = LabeledProperty(
        default=None, text="Linked with",
        visible=True, persistent=True,
        signal_name="data_changed",
        mix_with=(SignalMixin,)
    )
    @linked_with.setter
    def linked_with(self, value):
        old = type(self).linked_with._get(self)
        if old != value:
            if old is not None:
                self.relieve_model(old)
            type(self).linked_with._set(self, value)
            if value is not None:
                self.observe_model(value)
            else:
                for prop in self.Meta.get_inheritable_properties():
                    setattr(self, prop.inherit_flag, False)

    #: The silicate lattice's c length
    lattice_d = FloatProperty(
        default=0.0, text="Lattice c length [nm]",
        visible=False, persistent=True,
        signal_name="data_changed"
    )

    ucp_a = LabeledProperty(
        default=None, text="Cell length a [nm]",
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_ucp_a", inherit_from="linked_with.ucp_a",
        signal_name="data_changed",
        mix_with=(SignalMixin, InheritableMixin, ObserveMixin, RefinableMixin)
    )

    ucp_b = LabeledProperty(
        default=None, text="Cell length b [nm]",
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_ucp_b", inherit_from="linked_with.ucp_b",
        signal_name="data_changed",
        mix_with=(SignalMixin, InheritableMixin, ObserveMixin, RefinableMixin)
    )

    d001 = FloatProperty(
        default=1.0, text="Cell length c [nm]", minimum=0.0, maximum=5.0,
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_default_c", inherit_from="linked_with.d001",
        signal_name="data_changed",
        mix_with=(SignalMixin, InheritableMixin, RefinableMixin)
    )

    default_c = FloatProperty(
        default=1.0, text="Default c length [nm]", minimum=0.0, maximum=5.0,
        visible=True, persistent=True, tabular=True,
        inheritable=True, inherit_flag="inherit_default_c", inherit_from="linked_with.default_c",
        signal_name="data_changed",
        mix_with=(SignalMixin, InheritableMixin)
    )

    delta_c = FloatProperty(
        default=0.0, text="C length dev. [nm]", minimum=0.0, maximum=0.05,
        visible=True, persistent=True, tabular=True,
        inheritable=True, inherit_flag="inherit_delta_c", inherit_from="linked_with.delta_c",
        signal_name="data_changed",
        mix_with=(SignalMixin, InheritableMixin, RefinableMixin)
    )

    layer_atoms = ListProperty(
        default=None, text="Layer atoms",
        visible=True, persistent=True, tabular=True, widget_type="custom",
        inheritable=True, inherit_flag="inherit_layer_atoms", inherit_from="linked_with.layer_atoms",
        signal_name="data_changed", data_type=Atom,
        mix_with=(SignalMixin, InheritableMixin)
    )

    interlayer_atoms = ListProperty(
        default=None, text="Interlayer atoms",
        visible=True, persistent=True, tabular=True, widget_type="custom",
        inheritable=True, inherit_flag="inherit_interlayer_atoms", inherit_from="linked_with.interlayer_atoms",
        signal_name="data_changed", data_type=Atom,
        mix_with=(SignalMixin, InheritableMixin)
    )

    atom_relations = ListProperty(
        default=None, text="Atom relations", widget_type="custom",
        visible=True, persistent=True, tabular=True, refinable=True,
        inheritable=True, inherit_flag="inherit_atom_relations", inherit_from="linked_with.atom_relations",
        signal_name="data_changed", data_type=AtomRelation,
        mix_with=(SignalMixin, InheritableMixin, RefinableMixin)
    )

    # Instance flag indicating whether or not linked_with & inherit flags should be saved
    save_links = True
    # Class flag indicating whether or not atom types in the component should be
    # exported using their name rather then their project-uuid.
    export_atom_types = False

    # REFINEMENT GROUP IMPLEMENTATION:
    @property
    def refine_title(self):
        return self.name

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.phase.refine_title,
            component_name=self.refine_title
        )

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, **kwargs):
        """
        Valid keyword arguments for a Component are:
        *ucp_a*: unit cell property along a axis
        *ucp_b*: unit cell property along b axis
        *d001*: unit cell length c (aka d001)
        *default_c*: default c-value
        *delta_c*: the variation in basal spacing due to defects
        *layer_atoms*: ObjectListStore of layer Atoms
        *interlayer_atoms*: ObjectListStore of interlayer Atoms
        *atom_relations*: ObjectListStore of AtomRelations
        *inherit_ucp_a*: whether or not to inherit the ucp_a property from
         the linked component (if linked)
        *inherit_ucp_b*: whether or not to inherit the ucp_b property from
         the linked component (if linked)
        *inherit_d001*: whether or not to inherit the d001 property from
         the linked component (if linked)
        *inherit_default_c*: whether or not to inherit the default_c 
         property from the linked component (if linked)
        *inherit_delta_c*: whether or not to inherit the delta_c 
         property from the linked component (if linked)
        *inherit_layer_atoms*: whether or not to inherit the layer_atoms 
         property from the linked component (if linked)
        *inherit_interlayer_atoms*: whether or not to inherit the
         interlayer_atoms property from the linked component (if linked)
        *inherit_atom_relations*: whether or not to inherit the 
         atom_relations property from the linked component (if linked)
        *linked_with_uuid*: the UUID for the component this one is linked
         with
    Deprecated, but still supported:
        *linked_with_index*: the index of the component this one is 
         linked with in the ObjectListStore of the parent based on phase.
        """

        my_kwargs = self.pop_kwargs(kwargs,
            "data_name", "data_layer_atoms", "data_interlayer_atoms", "data_atom_relations",
            "data_atom_ratios", "data_d001", "data_default_c", "data_delta_c", "lattice_d",
            "data_cell_a", "data_ucp_a", "data_cell_b", "data_ucp_b",
            "linked_with_uuid", "linked_with_index", "inherit_cell_a", "inherit_cell_b",
            *[prop.label for prop in Component.Meta.get_local_persistent_properties()]
        )
        super(Component, self).__init__(**kwargs)
        kwargs = my_kwargs

        # Set up data object
        self._data_object = ComponentData(
            d001=0.0,
            delta_c=0.0
        )

        # Set attributes:
        self.name = self.get_kwarg(kwargs, "", "name", "data_name")

        # Load lists:
        self.layer_atoms = self.get_list(kwargs, [], "layer_atoms", "data_layer_atoms", parent=self)
        self.interlayer_atoms = self.get_list(kwargs, [], "interlayer_atoms", "data_interlayer_atoms", parent=self)
        self.atom_relations = self.get_list(kwargs, [], "atom_relations", "data_atom_relations", parent=self)

        # Add all atom ratios to the AtomRelation list
        for atom_ratio in self.get_list(kwargs, [], "atom_ratios", "data_atom_ratios", parent=self):
            self.atom_relations.append(atom_ratio)

        # Observe the inter-layer atoms, and make sure they get stretched
        for atom in self.interlayer_atoms:
            atom.stretch_values = True
            self.observe_model(atom)

        # Observe the layer atoms
        for atom in self.layer_atoms:
            self.observe_model(atom)

        # Resolve their relations and observe the atom relations
        for relation in self.atom_relations:
            relation.resolve_relations()
            self.observe_model(relation)

        # Connect signals to lists and dicts:
        self._layer_atoms_observer = ListObserver(
            self._on_layer_atom_inserted,
            self._on_layer_atom_removed,
            prop_name="layer_atoms",
            model=self
        )
        self._interlayer_atoms_observer = ListObserver(
            self._on_interlayer_atom_inserted,
            self._on_interlayer_atom_removed,
            prop_name="interlayer_atoms",
            model=self
        )
        self._atom_relations_observer = ListObserver(
            self._on_atom_relation_inserted,
            self._on_atom_relation_removed,
            prop_name="atom_relations",
            model=self
        )

        # Update lattice values:
        self.d001 = self.get_kwarg(kwargs, self.d001, "d001", "data_d001")
        self._default_c = float(self.get_kwarg(kwargs, self.d001, "default_c", "data_default_c"))
        self.delta_c = float(self.get_kwarg(kwargs, self.delta_c, "delta_c", "data_delta_c"))
        self.update_lattice_d()

        # Set/Create & observe unit cell properties:
        ucp_a = self.get_kwarg(kwargs, None, "ucp_a", "data_ucp_a", "data_cell_a")
        if isinstance(ucp_a, float):
            ucp_a = UnitCellProperty(name="cell length a", value=ucp_a, parent=self)
        ucp_a = self.parse_init_arg(
            ucp_a, UnitCellProperty, child=True,
            default_is_class=True, name="Cell length a [nm]", parent=self
        )
        type(self).ucp_a._set(self, ucp_a)
        self.observe_model(ucp_a)

        ucp_b = self.get_kwarg(kwargs, None, "ucp_b", "data_ucp_b", "data_cell_b")
        if isinstance(ucp_b, float):
            ucp_b = UnitCellProperty(name="cell length b", value=ucp_b, parent=self)
        ucp_b = self.parse_init_arg(
            ucp_b, UnitCellProperty, child=True,
            default_is_class=True, name="Cell length b [nm]", parent=self
        )
        type(self).ucp_b._set(self, ucp_b)
        self.observe_model(ucp_b)

        # Set links:
        self._linked_with_uuid = self.get_kwarg(kwargs, "", "linked_with_uuid")
        self._linked_with_index = self.get_kwarg(kwargs, -1, "linked_with_index")

        # Set inherit flags:
        self.inherit_d001 = self.get_kwarg(kwargs, False, "inherit_d001")
        self.inherit_ucp_a = self.get_kwarg(kwargs, False, "inherit_ucp_a", "inherit_cell_a")
        self.inherit_ucp_b = self.get_kwarg(kwargs, False, "inherit_ucp_b", "inherit_cell_b")
        self.inherit_default_c = self.get_kwarg(kwargs, False, "inherit_default_c")
        self.inherit_delta_c = self.get_kwarg(kwargs, False, "inherit_delta_c")
        self.inherit_layer_atoms = self.get_kwarg(kwargs, False, "inherit_layer_atoms")
        self.inherit_interlayer_atoms = self.get_kwarg(kwargs, False, "inherit_interlayer_atoms")
        self.inherit_atom_relations = self.get_kwarg(kwargs, False, "inherit_atom_relations")

    def __repr__(self):
        return "Component(name='%s', linked_with=%r)" % (self.name, self.linked_with)

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @DataModel.observe("data_changed", signal=True)
    def _on_data_model_changed(self, model, prop_name, info):
        # Check whether the changed model is an AtomRelation or Atom, if so
        # re-apply the atom_relations.
        with self.data_changed.hold():
            if isinstance(model, AtomRelation) or isinstance(model, Atom):
                self._apply_atom_relations()
                self._update_ucp_values()
            if isinstance(model, UnitCellProperty):
                self.data_changed.emit() # propagate signal

    @DataModel.observe("removed", signal=True)
    def _on_data_model_removed(self, model, prop_name, info):
        # Check whether the removed component is linked with this one, if so
        # clears the link and emits the data_changed signal.
        if model != self and self.linked_with is not None and self.linked_with == model:
            with self.data_changed.hold_and_emit():
                self.linked_with = None

    def _on_layer_atom_inserted(self, atom):
        """Sets the atoms parent and stretch_values property,
        updates the components lattice d-value, and emits a data_changed signal"""
        with self.data_changed.hold_and_emit():
            with self.atoms_changed.hold_and_emit():
                atom.parent = self
                atom.stretch_values = False
                self.observe_model(atom)
                self.update_lattice_d()

    def _on_layer_atom_removed(self, atom):
        """Clears the atoms parent, updates the components lattice d-value, and
        emits a data_changed signal"""
        with self.data_changed.hold_and_emit():
            with self.atoms_changed.hold_and_emit():
                self.relieve_model(atom)
                atom.parent = None
                self.update_lattice_d()

    def _on_interlayer_atom_inserted(self, atom):
        """Sets the atoms parent and stretch_values property, 
        and emits a data_changed signal"""
        with self.data_changed.hold_and_emit():
            with self.atoms_changed.hold_and_emit():
                atom.stretch_values = True
                atom.parent = self
    def _on_interlayer_atom_removed(self, atom):
        """Clears the atoms parent property, 
        and emits a data_changed signal"""
        with self.data_changed.hold_and_emit():
            with self.atoms_changed.hold_and_emit():
                atom.parent = None

    def _on_atom_relation_inserted(self, item):
        item.parent = self
        self.observe_model(item)
        self._apply_atom_relations()

    def _on_atom_relation_removed(self, item):
        self.relieve_model(item)
        item.parent = None
        self._apply_atom_relations()

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    def resolve_json_references(self):
        for atom in type(self).layer_atoms._get(self):
            atom.resolve_json_references()
        for atom in type(self).interlayer_atoms._get(self):
            atom.resolve_json_references()

        type(self).ucp_a._get(self).resolve_json_references()
        type(self).ucp_a._get(self).update_value()
        type(self).ucp_b._get(self).resolve_json_references()
        type(self).ucp_b._get(self).update_value()

        if getattr(self, "_linked_with_uuid", None):
            self.linked_with = type(type(self)).object_pool.get_object(self._linked_with_uuid)
            del self._linked_with_uuid
        elif getattr(self, "_linked_with_index", None) and self._linked_with_index != -1:
            warn("The use of object indeces is deprected since version 0.4. Please switch to using object UUIDs.", DeprecationWarning)
            self.linked_with = self.parent.based_on.components.get_user_from_index(self._linked_with_index)
            del self._linked_with_index

    @classmethod
    def save_components(cls, components, filename):
        """
            Saves multiple components to a single file.
        """
        Component.export_atom_types = True
        for comp in components:
            comp.save_links = False
        with zipfile.ZipFile(filename, 'w', compression=COMPRESSION) as zfile:
            for component in components:
                zfile.writestr(component.uuid, component.dump_object())
        for comp in components:
            comp.save_links = True
        Component.export_atom_types = False

        # After export we change all the UUID's
        # This way, we're sure that we're not going to import objects with
        # duplicate UUID's!
        type(cls).object_pool.change_all_uuids()

    @classmethod
    def load_components(cls, filename, parent=None):
        """
            Returns multiple components loaded from a single file.
        """
        # Before import, we change all the UUID's
        # This way we're sure that we're not going to import objects
        # with duplicate UUID's!
        type(cls).object_pool.change_all_uuids()
        if zipfile.is_zipfile(filename):
            with zipfile.ZipFile(filename, 'r') as zfile:
                for uuid in zfile.namelist():
                    obj = JSONParser.parse(zfile.open(uuid))
                    obj.parent = parent
                    yield obj
        else:
            obj = JSONParser.parse(filename)
            obj.parent = parent
            yield obj

    def json_properties(self):
        if self.phase == None or not self.save_links:
            retval = Storable.json_properties(self)
            for prop in self.Meta.all_properties:
                if getattr(prop, "inherit_flag", False):
                    retval[prop.inherit_flag] = False
        else:
            retval = Storable.json_properties(self)
            retval["linked_with_uuid"] = self.linked_with.uuid if self.linked_with is not None else ""
        return retval

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_factors(self, range_stl):
        """
        Get the structure factor for the given range of sin(theta)/lambda values.
        :param range_stl: A 1D numpy ndarray
        """
        return get_factors(range_stl, self.data_object)

    def get_interlayer_stretch_factors(self):
        z_factor = (self.cell_c - self.lattice_d) / (self.default_c - self.lattice_d)
        return self.lattice_d, z_factor

    def update_lattice_d(self):
        """
            Updates the lattice_d attribute for this :class:`~.Component`. 
            Should normally not be called from outside the component.
        """
        for atom in self.layer_atoms:
            self.lattice_d = float(max(self.lattice_d, atom.default_z))

    def _apply_atom_relations(self):
        """
        Applies the :class:`~..atom_relations.AtomRelation` objects
        in this component. Should normally not be called from outside the component.
        """
        with self.data_changed.hold_and_emit():
            for relation in self.atom_relations:
                # Clear the 'driven by' flags:
                relation.driven_by_other = False
            for relation in self.atom_relations:
                # Apply the relations, will also take care of flag setting:
                relation.apply_relation()

    def _update_ucp_values(self):
        """
        Updates the :class:`~..unit_cell_prop.UnitCellProperty` objects in this
        component. Should normally not be called from outside the component.
        """
        with self.data_changed.hold():
            for ucp in [self._ucp_a, self._ucp_b]:
                ucp.update_value()

    def get_volume(self):
        """
        Get the volume for this :class:`~.Component`.
        Will always return a value >= 1e-25, to prevent division-by-zero
        errors in calculation code.  
        """
        return max(self.cell_a * self.cell_b * self.cell_c, 1e-25)

    def get_weight(self):
        """
        Get the total atomic weight for this 
        :class:`~.Component`. 
        """
        weight = 0
        for atom in (self.layer_atoms + self.interlayer_atoms):
            weight += atom.weight
        return weight

    # ------------------------------------------------------------
    #      AtomRelation list related
    # ------------------------------------------------------------
    def move_atom_relation_up(self, relation):
        """
        Move the passed :class:`~..atom_relations.AtomRelation`
        up one slot
        """
        index = self.atom_relations.index(relation)
        del self.atom_relations[index]
        self.atom_relations.insert(max(index - 1, 0), relation)

    def move_atom_relation_down(self, relation):
        """
        Move the passed :class:`~..atom_relations.AtomRelation`
        down one slot
        """
        index = self.atom_relations.index(relation)
        del self.atom_relations[index]
        self.atom_relations.insert(min(index + 1, len(self.atom_relations)), relation)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .mixture import Mixture
#from .insitu_mixture import InSituMixture
#from .insitu_behaviours import InSituBehaviour

__all__ = [
    "Mixture",
    #"InSituMixture",
    #"InSituBehaviour"
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from mvc import Controller
from mvc.adapters.dummy_adapter import DummyAdapter

from pyxrd.generic.views import ChildObjectListStoreView
from pyxrd.generic.views.combobox_tools import add_combo_text_column
from pyxrd.generic.controllers import BaseController
from pyxrd.generic.controllers.objectliststore_controllers import wrap_list_property_to_treemodel

from pyxrd.probabilities.controllers import EditProbabilitiesController
from pyxrd.probabilities.views import EditProbabilitiesView

from pyxrd.phases.controllers import (
    EditCSDSTypeController, ComponentsController
)
from pyxrd.phases.views import EditCSDSDistributionView

class EditPhaseController(BaseController):
    """ 
        Controller for the phase edit view
    """
    probabilities_view = None
    probabilities_controller = None

    csds_view = None
    csds_controller = None

    components_view = None
    components_controller = None

    widget_handlers = {
        'custom': 'custom_handler',
    }

    @property
    def phases_treemodel(self):
        if self.model.project is not None:
            return wrap_list_property_to_treemodel(
                self.model.project, type(self.model.project).phases)
        else:
            return None

    def register_view(self, view):
        BaseController.register_view(self, view)

        self.csds_view = EditCSDSDistributionView(parent=self.view)
        self.view.set_csds_view(self.csds_view)

        if self.model.G > 1:
            self.probabilities_view = EditProbabilitiesView(parent=self.view)
            self.view.set_probabilities_view(self.probabilities_view)
        else:
            self.view.remove_probabilities()

        self.components_view = ChildObjectListStoreView(parent=self.view)
        self.components_view["button_add_object"].set_visible(False)
        self.components_view["button_add_object"].set_no_show_all(True)
        self.components_view["button_del_object"].set_visible(False)
        self.components_view["button_del_object"].set_no_show_all(True)
        self.view.set_components_view(self.components_view)

    @staticmethod
    def custom_handler(self, prop, widget): # TODO split out these 4 properties in their own adapters
        if prop.label in ("CSDS_distribution", "components", "probabilities", "based_on"):
            if prop.label == "CSDS_distribution":
                self.reset_csds_controller()
            elif prop.label == "components":
                self.reset_components_controller()
            elif prop.label == "probabilities":
                self.reset_probabilities_controller()
            elif prop.label == "based_on" and self.phases_treemodel is not None:
                combo = self.view["phase_based_on"]

                combo.set_model(self.phases_treemodel)
                combo.connect('changed', self.on_based_on_changed)

                def phase_renderer(celllayout, cell, model, itr, user_data=None):
                    phase = model.get_user_data(itr)
                    if phase: # FIXME an error can occur here if the phase list is cleared and the view is still open
                        cell.set_sensitive(phase.R == self.model.R and phase.G == self.model.G and phase.get_based_on_root() != self.model)
                add_combo_text_column(combo, data_func=phase_renderer, text_col=self.phases_treemodel.c_name)

                for row in self.phases_treemodel:
                    if self.phases_treemodel.get_user_data(row.iter) == self.model.based_on:
                        combo.set_active_iter (row.iter)
                        break

            return DummyAdapter(controller=self, prop=prop)

    def reset_csds_controller(self):
        if self.csds_controller is None:
            self.csds_controller = EditCSDSTypeController(
                model=self.model, view=self.csds_view, parent=self)
        else:
            self.csds_controller.model = self.model

    def reset_components_controller(self):
        self.components_controller = ComponentsController(
            model=self.model, view=self.components_view, parent=self)

    def reset_probabilities_controller(self):
        if self.probabilities_controller is None:
            if self.model.G > 1: # False if model is a multi-component phase
                self.probabilities_controller = EditProbabilitiesController(
                    model=self.model.probabilities,
                    view=self.probabilities_view, parent=self)
        else:
            self.probabilities_controller.model = self.model.probabilities

    def register_adapters(self):
        self.update_sensitivities()

    def update_sensitivities(self):
        can_inherit = (self.model.based_on is not None)

        for name in ("sigma_star", "display_color"):
            widget_name = "container_%s" % name
            self.view[widget_name].set_sensitive(not (can_inherit and getattr(self.model, "inherit_%s" % name)))
            self.view[widget_name].set_visible(not (can_inherit and getattr(self.model, "inherit_%s" % name)))
            self.view["phase_inherit_%s" % name].set_sensitive(can_inherit)

        for name in ("CSDS_distribution",):
            sensitive = not (can_inherit and getattr(self.model, "inherit_%s" % name))
            self.view["phase_inherit_%s" % name].set_sensitive(can_inherit)
            self.view.set_csds_sensitive(sensitive)
            self.reset_csds_controller()

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("inherit_display_color", assign=True)
    @Controller.observe("inherit_sigma_star", assign=True)
    @Controller.observe("inherit_CSDS_distribution", assign=True)
    def notif_change_inherit(self, model, prop_name, info):
        self.update_sensitivities()
        return

    @Controller.observe("probabilities", assign=True)
    def notif_change_probabilities(self, model, prop_name, info):
        self.reset_probabilities_controller()
        return

    @Controller.observe("name", assign=True)
    def notif_name_changed(self, model, prop_name, info):
        self.phases_treemodel.on_item_changed(self.model)
        return

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_based_on_changed(self, combo, user_data=None):
        itr = combo.get_active_iter()
        if itr is not None:
            val = combo.get_model().get_user_data(itr)
            # cannot be based on itself == not based on anything
            # cannot be based on a model with a different # of components
            if val != self.model and val.get_based_on_root() != self.model and val.G == self.model.G:
                self.model.based_on = val
                self.update_sensitivities()
                return
        combo.set_active(-1)
        self.update_sensitivities()
        self.model.based_on = None



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import weakref
from .base import Observer

class ListItemObserver(Observer):
    """
        An observer that observes a single item in a list and informs us of changes.
        The observed properties are defined in the list type's meta class by
        setting their property descriptors 'tabular' attribute to True.
    """

    _previous_model_ref = None
    @property
    def _previous_model(self):
        if self._previous_model_ref is not None:
            return self._previous_model_ref()
        else:
            return None
    @_previous_model.setter
    def _previous_model(self, value):
        self._previous_model_ref = weakref.ref(value, self.clear)

    def __init__(self, on_changed, model=None, spurious=False):
        super(ListItemObserver, self).__init__(spurious=spurious)
        self.on_changed = on_changed
        self.observe_model(model)

    def observe_model(self, model):
        if self._previous_model is not None:
            self.relieve_model(self._previous_model)
        if model is not None:
            for prop_name, data_type in model.Meta.get_column_properties():  # @UnusedVariable
                self.observe(self.on_prop_mutation, prop_name, assign=True)
            self._previous_model = model
            super(ListItemObserver, self).observe_model(model)

    def clear(self, *args):
        self.on_changed = None
        if len(args) == 0:
            self.observe_model(None)

    def on_prop_mutation(self, model, prop_name, info):
        if callable(self.on_changed):
            self.on_changed(model)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .labeled_property import LabeledProperty

class CastProperty(LabeledProperty):
    """
     A descriptor that can cast the values to a given type and clamp values to a
     minimum and maximum. 
     Expects its label to be set or passed to __init__.
    """

    def __init__(self,
            minimum=None, maximum=None,
            cast_to=None,
            *args, **kwargs):
        super(CastProperty, self).__init__(*args, **kwargs)
        self.minimum = minimum
        self.maximum = maximum
        self.cast_to = cast_to

    def __cast_and_clamp__(self, instance, value):
        if self.minimum is not None:
            value = max(value, self.minimum)
        if self.maximum is not None:
            value = min(value, self.maximum)
        if self.cast_to is not None and value is not None:
            value = self.cast_to(value)
        return value

    def __set__(self, instance, value):
        value = self.__cast_and_clamp__(instance, value)
        if getattr(instance, self.label) != value:
            super(CastProperty, self).__set__(instance, value)

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

def _toggle_cb(dialog, event, cb):
    cb_id_name = "%s_cb_id" % event.replace("-", "_")
    cb_id = getattr(dialog, cb_id_name, None)
    if cb_id is not None: dialog.disconnect(cb_id)
    cb_id = dialog.connect(event, cb)
    setattr(dialog, cb_id_name, cb_id)

def run_dialog(dialog,
        on_accept_callback=None, on_reject_callback=None, destroy=True):
    """ Helper method - do not call directly """

    if not (on_accept_callback is None or callable(on_accept_callback)):
        raise ValueError("Accept callback must be None or callable")
    if not (on_reject_callback is None or callable(on_reject_callback)):
        raise ValueError("Reject callback must be None or callable")      
   
    def _dialog_response_cb(dialog, response): 
        if response in dialog.accept_responses and on_accept_callback is not None:
            on_accept_callback(dialog)
        elif on_reject_callback is not None:
            on_reject_callback(dialog)
        if destroy:
            dialog.destroy()
        else:
            dialog.hide()
        return not destroy
    _toggle_cb(dialog, "response", _dialog_response_cb)
    
    # Adding the delete event prevents the dialog from being destroyed if the
    # user indicated it should persist
    def delete_event_cb(dialog, event):
        if on_reject_callback is not None:
            on_reject_callback(dialog)
        if destroy:
            dialog.destroy()
        else:
            dialog.hide()
        return not destroy
    _toggle_cb(dialog, "delete_event", delete_event_cb)
    
    # Present the dialog
    dialog.set_modal(True)
    dialog.show_all()

def retrieve_lowercase_extension(glob):
    '''Ex: '*.[oO][rR][aA]' => '*.ora' '''
    return ''.join([ c.replace("[", "").replace("]", "")[:-1] for c in glob.split('][')])

def adjust_filename_to_globs(filename, globs):
    """ Adjusts a given filename so it ends with the proper extension """
    if globs: # If given use file extensions globs
        possible_fns = []
        # Loop over globs, if the current filenames extensions matches
        # a given glob, the filename is returned as is, otherwise
        # the extension of the first glob is added at the end of filename
        for glob in globs:
            if glob is not None:
                extension = glob[1:]
                if filename[len(filename) - len(extension):].lower() != extension.lower():
                    possible_fns.append("%s%s" % (filename, glob[1:]))
                else:
                    return filename # matching extension is returned immediately
        return possible_fns[0] # otherwise add extension of the first filter
    else: # If no globs are given, return filename as is
        return filename




# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import importlib
import logging
logger = logging.getLogger(__name__)

from ..settings import TOOLKIT
from ..support import gui_loop

class ToolkitRegistry(dict):
    """
        Dict subclass used to store all AdapterRegistry's.
        Keys are toolkit names, values are AdapterRegistry instances
    """
    
    # These will be loaded automatically when this module is first loaded:
    toolkit_modules = [
        ".gtk_support"
    ]
    selected_toolkit = None

    def get_or_create_registry(self, toolkit_name):
        if not toolkit_name in self:
            adapter_registry = AdapterRegistry()
            self.register(toolkit_name, adapter_registry)
        return self[toolkit_name]

    def register(self, toolkit_name, adapter_registry):
        self[toolkit_name] = adapter_registry

    def select_toolkit(self, toolkit_name):
        if not toolkit_name in self:
            raise ValueError("Cannot select unknown toolkit '%s'" % toolkit_name)
        else:
            self.selected_toolkit = toolkit_name
            tkar = self.get_selected_adapter_registry()
            tkar.load_toolkit_functions()
            
    def get_selected_adapter_registry(self):
        if self.selected_toolkit is None:
            raise ValueError("No toolkit has been selected!")
        else:
            return self[self.selected_toolkit]

class AdapterRegistry(dict):
    """
        A dict which maps Adapter class types to the widget types they
        can handle. This relies on these classes being registered using
        the 'register' decorator also provided by this class.
    """

    toolkit_registry = ToolkitRegistry()

    @classmethod
    def get_selected_adapter_registry(cls):
        return cls.toolkit_registry.get_selected_adapter_registry()

    @classmethod
    def get_adapter_for_widget_type(cls, widget_type):
        return cls.toolkit_registry.get_selected_adapter_registry()[widget_type]

    def set_toolkit_functions(self, *args, **kwargs):
        setattr(self, "_toolkit_functions", (args, kwargs))

    def load_toolkit_functions(self):
        """
            This function loads the toolkit functions passed to 
            set_toolkit_functions with the support module.
        """
        args, kwargs = getattr(self, "_toolkit_functions", ([],{}))
        gui_loop.load_toolkit_functions(*args, ** kwargs)

    @classmethod
    def register(cls, adapter_cls):
        """
            This is called from metaclasses or used as a decorator.
            An example metaclass is at mvc.adapters.metaclasses and a model
            implementing it at adapters.gtk_support.basic
        """
        if hasattr(adapter_cls, "widget_types") and hasattr(adapter_cls, "toolkit"):
            adapter_registry = cls.toolkit_registry.get_or_create_registry(adapter_cls.toolkit)
            logger.debug("Registering %s as handler for widget types '%s' in toolkit '%s'" % (adapter_cls, adapter_cls.widget_types, adapter_cls.toolkit))
            for widget_type in adapter_cls.widget_types:
                adapter_registry[widget_type] = adapter_cls
        else:
            logger.debug("Ignoring '%s' as handler: no 'toolkit' or 'widget_types' defined" % adapter_cls)
        return adapter_cls

    pass # end of class

for toolkit_module in ToolkitRegistry.toolkit_modules:
    if toolkit_module.startswith('.'):
        package = __name__.rpartition('.')[0]
        try:
            # Import and load toolkit module:
            tk_mod = importlib.import_module(toolkit_module, package=package)
            # Get or create the associated toolkit registry:
            tkreg = AdapterRegistry.toolkit_registry.get_or_create_registry(tk_mod.toolkit)
            # Load the toolkit and pass the registry:
            tk_mod.load(tkreg)
        except ImportError:
            logger.warning("Could not load toolkit support module '%s'" % (toolkit_module,))

AdapterRegistry.toolkit_registry.select_toolkit(TOOLKIT)



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

from pyxrd.generic.views import BaseView, DialogView

class InlineGoniometerView(BaseView):
    """
        The inline Goniometer view.
    """

    builder = resource_filename(__name__, "glade/goniometer.glade")
    top = "edit_goniometer"

    widget_format = "gonio_%s"

    @property
    def import_combo_box(self):
        return self["cmb_import_gonio"]

    pass # end of class


class WavelengthDistributionView(DialogView):
    """
        The wavelength distribution view.
    """   
    subview_builder = resource_filename(__name__, "glade/wavelength_distribution.glade")
    subview_toplevel = "edit_wld"

    widget_format = "wld_%s"
    
    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import csv, os

from pyxrd.generic.io import get_case_insensitive_glob
from pyxrd.generic.utils import u

from pyxrd.file_parsers.csv_base_parser import CSVBaseParser

from .namespace import atom_type_parsers

@atom_type_parsers.register_parser()
class ATLAtomTypeParser(CSVBaseParser):
    """
        Atomic scattering factors CSV file parser
    """

    namespace = "atl"
    description = "Atom types CSV file"
    extensions = get_case_insensitive_glob("*.ATL")

    default_fmt_params = {
        "delimiter": ',',
        "doublequote": True,
        "escapechar": None,
        "quotechar": "\"",
        "quoting": csv.QUOTE_MINIMAL,
        "skipinitialspace": True,
        "strict": False
    }

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False, **fmt_params):
        try:
            fmt_params = dict(cls.default_fmt_params, **fmt_params)
    
            # Goto start of file
            fp.seek(0)
    
            # Get base filename:
            basename = u(os.path.basename(filename))
    
            # Read in the first and last data line and put the file cursor back
            # at its original position
            header = cls.parse_raw_line(fp.readline().strip(), str)
            replace = [
                "a%d" % i for i in range (1, 6)
            ] + [
                "b%d" % i for i in range (1, 6)
            ] + [ "c", ]
            header = [ "par_%s" % val if val in replace else val for val in header ]
            data_start_pos = fp.tell()
            line_count, _ = cls.get_last_line(fp)
            fp.seek(data_start_pos)
    
            # Adapt DataObject list
            data_objects = cls._adapt_data_object_list(data_objects, num_samples=line_count)
    
            # Fill in header info:
            for i in range(line_count):
                data_objects[i].update(
                    filename=basename,
                    header=header
                )
        finally:
            if close: fp.close()
        return data_objects

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False, **fmt_params):
        if fp is not None:
            for row, data_object in zip(csv.reader(fp, **fmt_params), data_objects):
                if row:
                    for key, val in zip(data_object.header, row):
                        setattr(data_object, key, val)
                    data_object.is_json = False

        if close: fp.close()
        return data_objects

    @classmethod
    def parse(cls, filename, fp, data_objects=None, close=True, **fmt_params):
        """
            Files are sniffed for the used csv dialect, but an optional set of
            formatting parameters can be passed that will override the sniffed
            parameters.
        """
        filename, fp, close = cls._get_file(filename, fp, close=close)

        # Guess dialect
        fmt_params, _, _ = cls.sniff(fp, **fmt_params)

        # Parse header
        data_objects = cls.parse_header(filename, fp, data_objects=data_objects, **fmt_params)

        # Parse data
        data_objects = cls.parse_data(filename, fp, data_objects=data_objects, **fmt_params)

        if close: fp.close()
        return data_objects

    @classmethod
    def write(cls, filename, items, props):
        """
            Writes the header to the first line, and will write x, y1, ..., yn
            rows for each column inside the x and ys arguments.
            Header argument should not include a newline, and can be a string or
            any iterable containing strings.
        """
        atl_writer = csv.writer(open(filename, 'wb'), delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        atl_writer.writerow([ name for prop, name in props])
        for item in items:
            prop_row = []
            for prop, name in props:
                prop_row.append(getattr(item, prop))
            atl_writer.writerow(prop_row)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from traceback import format_exc
import itertools
from math import sqrt

import logging
from pyxrd.data import settings
logger = logging.getLogger(__name__)

import numpy as np
from scipy.interpolate.ndgriddata import griddata

from mpl_toolkits.axes_grid1 import ImageGrid
import mpl_toolkits.axes_grid1.axes_size as Size
from mpl_toolkits.axes_grid1 import Divider
from pyxrd.generic.mathtext_support import get_plot_safe

class ParameterSpaceGenerator(object):

    num_cross_sections = 0
    grid = None

    def initialize(self, ranges, density):
        """
            Create a new generator with the given minimum and maximum values
            and the given grid density (expressed as # of data points for that 
            parameter axis).
        """

        self.num_params = len(ranges)
        assert self.num_params >= 1, "Need to have at least one refinable parameter!"

        ranges = np.asarray(ranges, dtype=float)

        self.minima = np.asarray(ranges[:, 0])
        self.maxima = np.asarray(ranges[:, 1])
        self.center_point = 0.5 * (self.minima + self.maxima)

        self.num_cross_sections = int(self.num_params * (self.num_params + 1) / 2)
        self.grid_dtype = np.dtype([("point", object), ("value", float), ("distance", float)])
        self.density = self._calculate_density(density)

        if self.num_params == 1:
            self.grid = np.empty(shape=(self.density,), dtype=self.grid_dtype)
        else:
            self.grid = np.empty(shape=(self.num_cross_sections, self.density, self.density), dtype=self.grid_dtype)

    def _calculate_density(self, density, memory_limit=settings.PAR_SPACE_MEMORY_LIMIT):
        block_size = self.grid_dtype.itemsize
        limited = sqrt(memory_limit / (block_size * self.num_cross_sections))
        return max(min(density, limited), 3)

    def _find_closest_grid_index(self, solution):
        """ Returns the closest grid point's indexes """

        # Transform solution so center of grid is zero point and min-max goes from -1 to +1
        transf = (np.asarray(solution) - self.center_point) / (self.maxima - self.minima)

        if self.num_params > 1:
            # Calculate distance to each cross section and find closest
            smallest_distance = None
            closest_cross_section = None

            for index, (par1, par2) in enumerate(itertools.combinations(list(range(self.num_params)), 2)):
                # Project on to the normal of the A/B plan at the center point:
                projected = transf.copy()
                projected[par1] = 0
                projected[par2] = 0
                distance = np.linalg.norm(projected)
                if smallest_distance is None or smallest_distance > distance:
                    smallest_distance = distance
                    closest_cross_section = index, (par1, par2), distance

            index, (par1, par2), distance = closest_cross_section

            # Move back to 0->density space (was in -1 -> 1 space)
            gridded_location = np.array(np.round(0.5 * (transf + 1) * (self.density - 1)), dtype=int)
            closest_index = (index, int(gridded_location[par1]), int(gridded_location[par2]))
        else:
            # Only a single parameter, so only a single index:
            gridded_location = 0.5 * (transf + 1.0) * (self.density - 1.0)
            distance = np.abs(np.round(gridded_location) - gridded_location)
            closest_index = (int(np.round(gridded_location)[0]),)

        return closest_index, distance

    total_record_calls = 0
    total_actual_records = 0

    def record(self, new_solution, new_residual):
        """
            Add a new solution to the list of solutions
        """
        self.total_record_calls += 1
        # Get the best spot to store this one:
        closest_index, new_distance = self._find_closest_grid_index(new_solution)

        old_item = self.grid[closest_index]
        old_solution = old_item["point"]
        #old_residual = old_item["value"]
        old_distance = old_item["distance"]

        # If we have a previous point, check which one is closer:
        if (old_solution is not None and old_solution[0] is not None) and new_distance > old_distance:
            return
        else:
            self.total_actual_records += 1
            # If we got here, we can store the result:
            self.grid[closest_index] = (new_solution, new_residual, new_distance)

    def clear(self):
        del self.grid

    def clear_image(self, figure, message="Interpolation Error"):
        figure.clear()
        figure.text(0.5, 0.5, message, va="center", ha="center")

    def _setup_image_grid(self, figure, dims):
        """
            An example of how grid, parameter and view numbers change
            for dims = 4
            
            The numbers in the grid are:
            
            parameter x, parameter y
            grid x, grid y
        
            -----------------------------------------------------
            |            |            |            |            |
            |    0, 0    |    1, 0    |    2, 0    |    3, 0    |
            |    -, -    |    -, -    |    -, -    |    -, -    |
            |            |            |            |            |
            ==============------------|------------|------------|
            I            I            |            |            |
            I    0, 1    I    1, 1    |    2, 1    |    3, 1    |
            I    0, 0    I    1, 0    |    2, 0    |    -, -    |
            I            I            |            |            |
            I------------==============------------|------------|
            I            |            I            |            |
            I    0, 2    |    1, 2    I    2, 2    |    3, 2    |
            I    0, 1    |    1, 1    I    2, 1    |    -, -    |
            I            |            I            |            |
            I------------|------------==============------------|
            I            |            |            I            |
            I    0, 3    |    1, 3    |    2, 3    I    3, 3    |
            I    0, 2    |    1, 2    |    2, 2    I    -, -    |
            I            |            |            I            |
            I======================================I------------|
            From the above it should be clear that:
            
            parameter x = grid x
            parameter y = grid y + 1
            grid nr = grid y + grid x * (dims - 1)
            view nr = grid nr - (grid nr / dims) * ((grid nr / dims) +1) / 2
            

        """
        image_grid = ImageGrid(
            figure, 111,
            nrows_ncols=(dims - 1, dims - 1),
            cbar_location="right",
            cbar_mode="single",
            # add_all=False,
            aspect=False,
            axes_pad=0.1,
            direction="column"
        )

        rect = (0.1, 0.1, 0.8, 0.8)
        horiz = [Size.Fixed(.1)] + [Size.Scaled(1.), Size.Fixed(.1)] * max(dims - 1, 1) + [Size.Fixed(0.15)]
        vert = [Size.Fixed(.1)] + [Size.Scaled(1.), Size.Fixed(.1)] * max(dims - 1, 1)

        # divide the axes rectangle into grid whose size is specified by horiz * vert
        divider = Divider(figure, rect, horiz, vert) # , aspect=False)

        # Helper to get the axis for par x and y:
        def get_grid(parx, pary):
            gridx = parx
            gridy = pary - 1
            return image_grid[gridy + gridx * (dims - 1)]

        # Helper to get the grid locator for par x and par y
        def get_locator(parx, pary):
            gridx = parx
            gridy = pary - 1
            nx = 1 + gridx * 2
            ny = 1 + (dims - gridy - 2) * 2
            return divider.new_locator(nx=nx, ny=ny)

        # Hide the unused plots & setup the used ones:
        for parx, pary in itertools.product(list(range(self.num_params - 1)), list(range(1, self.num_params))):

            # Calculate the grid position:
            ax = get_grid(parx, pary)

            # Setup axes:
            if pary <= parx:
                ax.set_visible(False)
            else:
                ax.set_axes_locator(get_locator(parx, pary))
                ax.set_visible(True)

        return image_grid, divider, get_grid

    def plot_images(self, figure, centroid, labels, density=200, smooth=0.5):
        """
            Generate the parameter space plots
        """

        try:

            def extraxt_points_from_grid(grid2D, parx=None, pary=None):
                """ Helper function that extract x,y(,z) points from all the data """
                grid2D = grid2D.flatten()
                if parx is not None and pary is not None:
                    points_x = np.array([item["point"][parx] for item in grid2D if item["point"] is not None])
                    points_y = np.array([item["point"][pary] for item in grid2D if item["point"] is not None])
                    points_z = np.array([item["value"] for item in grid2D if item["point"] is not None])
                    return points_x, points_y, points_z
                else:
                    points_x = np.array([item["point"] for item in grid2D if item["point"] is not None])
                    points_y = np.array([item["value"] for item in grid2D if item["point"] is not None])
                    points_x = points_x.flatten()
                    points_y = points_y.flatten()
                    xy = np.array(list(zip(points_x, points_y)), dtype=[('x', float), ('y', float)])
                    xy.sort(order=['x'], axis=0)
                    points_x = xy['x']
                    points_y = xy['y']
                    return points_x, points_y

            if self.num_params == 1:
                # Only one parameter refined:
                points_x, points_y = extraxt_points_from_grid(self.grid)
                ax = figure.add_subplot(1, 1, 1)
                ax.plot(points_x, points_y)
                ax.set_ylabel("Residual error")
                ax.set_xlabel(get_plot_safe(labels[0]))

            else:
                # Multi-parameter space:
                image_grid, divider, get_grid = self._setup_image_grid(figure, self.num_params)

                # Keep a reference to the images created,
                # so we can add a scale bar for all images (and they have the same range)
                ims = []
                tvmin, tvmax = None, None

                for index, (parx, pary) in enumerate(itertools.combinations(list(range(self.num_params)), 2)):
                    # Get the data for this cross section:
                    grid2D = self.grid[index, ...]
                    points_x, points_y, points_z = extraxt_points_from_grid(grid2D, parx, pary)

                    # Setup axis:
                    ax = get_grid(parx, pary)
                    extent = (
                        self.minima[parx], self.maxima[parx],
                        self.minima[pary], self.maxima[pary]
                    )
                    aspect = 'auto'

                    # Try to interpolate the data:
                    xi = np.linspace(self.minima[parx], self.maxima[parx], density)
                    yi = np.linspace(self.minima[pary], self.maxima[pary], density)
                    zi = griddata((points_x, points_y), points_z, (xi[None, :], yi[:, None]), method='cubic')

                    # Plot it:
                    im = ax.imshow(zi, origin='lower', aspect=aspect, extent=extent, alpha=0.75)
                    ims.append(im)
                    im.set_cmap('gray_r')

                    # Get visual limits:
                    vmin, vmax = im.get_clim()
                    if index == 0:
                        tvmin, tvmax = vmin, vmax
                    else:
                        tvmin, tvmax = min(tvmin, vmin), max(tvmax, vmax)

                    # Try to add a contour & labels:
                    try:
                        ct = ax.contour(xi, yi, zi, colors='k', aspect=aspect, extent=extent, origin='lower')
                        ax.clabel(ct, colors='k', fontsize=10, format="%1.2f")
                    except ValueError:
                        pass #ignore the "zero-size array" error for now.

                    # Add a red cross where the 'best' solution is:
                    ax.plot((centroid[parx],), (centroid[pary],), 'r+')

                    # Rotate x labels
                    for lbl in ax.get_xticklabels():
                        lbl.set_rotation(90)

                    # Reduce number of ticks:
                    ax.locator_params(axis='both', nbins=5)

                    # Set limits:
                    ax.set_xlim(extent[0:2])
                    ax.set_ylim(extent[2:4])

                    # Add labels to the axes so the user knows which is which:
                    # TODO add some flags/color/... indicating which phase & component each parameter belongs to...
                    if parx == 0:
                        ax.set_ylabel(str("#%d " % (pary + 1)) + get_plot_safe(labels[pary]))
                    if pary == (self.num_params - 1):
                        ax.set_xlabel(str("#%d " % (parx + 1)) + get_plot_safe(labels[parx]))

                # Set the data limits:
                for im in ims:
                    im.set_clim(tvmin, tvmax)

                # Make it look PRO:
                if im is not None:
                    cbar_ax = image_grid.cbar_axes[0]
                    nx = 1 + (self.num_params - 1) * 2
                    ny1 = (self.num_params - 1) * 2
                    cbar_ax.set_axes_locator(divider.new_locator(nx=nx, ny=1, ny1=ny1))
                    cb = cbar_ax.colorbar(im) # @UnusedVariable

        except:
            print("Unhandled exception while generating parameter space images:")
            print(format_exc())
            # ignore error, tell the user via the plot and return
            self.clear_image(figure)
            return

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (c) 2007 by Guillaume Libersat <glibersat AT linux62.org>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

# TODO remove gladeXML: deprecated!

from .support.exceptions import ViewError

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
Builder = Gtk.Builder

# ----------------------------------------------------------------------

class View (object):
    top = None
    builder = None

    def __init__(self, top=None,
                 parent=None,
                 builder=None, *args, **kwargs):
        """
        Only the first three may be given as positional arguments. If an
        argument is empty a class attribute of the same name is used. This
        does not work for *parent*.

        *builder* is a path to a Glade XML file.

        *top* is a string containing the name of our top level widget.

        *parent* is used to call :meth:`set_parent_view`.

        The last two only work if *builder* is used, not if you
        intend to create widgets later from code.
        """
        super(View, self).__init__(*args, **kwargs)

        self.manualWidgets = {}
        self.autoWidgets = {}
        self.__autoWidgets_calculated = False

        if top: self._top = top
        else: self._top = self.top

        # retrieves objects from builder if available
        if builder: _builder = builder
        else: _builder = self.builder
        if _builder is not None:
            # if the user passed a Builder, use it as it is, otherwise
            # build one
            if isinstance(_builder, Builder):
                self._builder = _builder
            else:
                self._builder = Gtk.Builder()
                self._builder.add_from_file(_builder)
                pass
            pass
        else: self._builder = None # no gtk builder

        if parent is not None: self.set_parent_view(parent)

        return

    def __getitem__(self, key):
        """
        Return the widget named *key*, or ``None``.
        
        .. note::
        
           In future versions this will likely change to raise ``KeyError``.
        """
        wid = None

        # first try with manually-added widgets:
        if key in self.manualWidgets:
            wid = self.manualWidgets[key]
            pass

        if wid is None:
            # then try with glade and builder, starting from memoized
            if key in self.autoWidgets: wid = self.autoWidgets[key]
            else:
                # try with Gtk.builder
                if wid is None and self._builder is not None:
                    wid = self._builder.get_object(key)
                    if wid is not None:
                        self.autoWidgets[key] = wid
                        pass
                    pass
                pass
            pass

        return wid

    def __setitem__(self, key, wid):
        """
        Add a widget. This overrides widgets of the same name that were loaded
        from XML. It does not affect GTK container/child relations.
        
        If no top widget is known, this sets it.
        """
        self.manualWidgets[key] = wid
        return

    def show(self):
        """
        Call `show()` on each top widget or `show_all()` if only one is known. 
        Otherwise does nothing.
        """
        top = self.get_top_widget()
        if type(top) in (list, tuple):
            for t in top:
                if t is not None: t.show()
                pass
        elif (top is not None): top.show_all()
        return


    def hide(self):
        """
        Call `hide()` on all known top widgets.
        """
        top = self.get_top_widget()
        if type(top) in (list, tuple):
            for t in top:
                if t is not None: t.hide()
                pass
        elif top is not None: top.hide()
        return

    def get_top_widget(self):
        return self[self._top]

    def set_parent_view(self, parent_view):
        """
        Set ``self.``:meth:`get_top_widget` transient for 
        ``parent_view.get_top_widget()``.
        """
        top = self.get_top_widget()
        if type(top) in (list, tuple):
            for t in top:
                if t is not None and hasattr(t, "set_transient_for"):
                    t.set_transient_for(parent_view.get_top_widget())
                    pass
                pass
        elif (top is not None) and hasattr(top, "set_transient_for"):
            top.set_transient_for(parent_view.get_top_widget())
            pass

        return

    def set_transient(self, transient_view):
        """
        Set ``transient_view.get_top_widget()`` transient for
        ``self.``:meth:`get_top_widget`.
        """
        top = self.get_top_widget()
        if type(top) in (list, tuple):
            for t in top:
                if t is not None:
                    transient_view.get_top_widget().set_transient_for(t)
                    pass
                pass
        elif (top is not None):
            transient_view.get_top_widget().set_transient_for(top)
            pass
        return

    # Finds the right callback for custom widget creation and calls it
    # Returns None if an undefined or invalid  handler is found
    def _custom_widget_create(self, glade, function_name, widget_name,
                              str1, str2, int1, int2):
        # This code was kindly provided by Allan Douglas <zalguod at
        # users.sourceforge.net>
        if function_name is not None:
            handler = getattr(self, function_name, None)
            if handler is not None: return handler(str1, str2, int1, int2)
            pass
        return None

    def __iter__(self):
        """
        Return an iterator over widgets added with :meth:`__setitem__` and
        those loaded from XML.
        
        .. note::
           In case of name conflicts this yields widgets that are not 
           accessible via :meth:`__getitem__`.
        """
        # precalculates if needed
        self.__extract_autoWidgets()

        import itertools
        for i in itertools.chain(self.manualWidgets, self.autoWidgets): yield i
        return

    def __extract_autoWidgets(self):
        """Extract autoWidgets map if needed, out of the glade
        specifications and gtk builder"""
        if self.__autoWidgets_calculated: return

        if self._builder is not None:
            for wid in self._builder.get_objects():
                # General workaround for issue
                # https://bugzilla.gnome.org/show_bug.cgi?id=607492
                try: name = Gtk.Buildable.get_name(wid)
                except TypeError: continue

                if name in self.autoWidgets and self.autoWidgets[name] != wid:
                    raise ViewError("Widget '%s' in builder also found in glade specification" % name)

                self.autoWidgets[name] = wid
                pass
            pass

        self.__autowidgets_calculated = True
        return

    pass # end of class View



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .insitu_behaviour import InSituBehaviour
from .SSSR0Behaviour import SSSR0Behaviour

__all__ = [
    "InSituBehaviour",
    "SSSR0Behaviour"
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
logger = logging.getLogger(__name__)

from mvc import Controller
from mvc.adapters.dummy_adapter import DummyAdapter

from pyxrd.generic.views import InlineObjectListStoreView
from pyxrd.generic.views.combobox_tools import add_combo_text_column
from pyxrd.generic.controllers import BaseController, ChildObjectListStoreController
from pyxrd.generic.io.utils import get_case_insensitive_glob

from pyxrd.phases.controllers import EditLayerController, EditAtomRelationsController, EditUnitCellPropertyController
from pyxrd.phases.views import EditComponentView, EditUnitCellPropertyView
from pyxrd.phases.models import Phase, Component
from pyxrd.generic.controllers.objectliststore_controllers import wrap_list_property_to_treemodel

class EditComponentController(BaseController):
    """ 
        Controller for the component edit view
    """
    layer_view = None
    layer_controller = None

    interlayer_view = None
    interlayer_controller = None

    atom_relations_view = None
    atom_relations_controller = None

    ucpa_view = None
    ucpa_controller = None

    ucpb_view = None
    ucpb_controller = None

    widget_handlers = {
        'custom': 'custom_handler',
    }

    @property
    def components_treemodel(self):
        if self.model.phase.based_on:
            return wrap_list_property_to_treemodel(
                self.model.phase.based_on, Phase.components
            )
        else:
            return None

    def reset_combo_box(self):
        """
            Reset the `linked_with` combo box.
        """
        if self.model is not None and self.model.parent is not None:
            combo = self.view["component_linked_with"]
            combo.clear()
            combo.set_model(self.components_treemodel)
            if self.components_treemodel is not None:
                add_combo_text_column(combo, text_col=self.components_treemodel.c_name)
                for row in self.components_treemodel:
                    comp = self.components_treemodel.get_user_data(row.iter)
                    if comp == self.model.linked_with:
                        combo.set_active_iter (row.iter)
                        break

    @staticmethod
    def custom_handler(self, prop, widget):
        if prop.label == "layer_atoms":
            self.view.set_layer_view(self.layer_view.get_top_widget())
        elif prop.label == "interlayer_atoms":
            self.view.set_interlayer_view(self.interlayer_view.get_top_widget())
        elif prop.label == "atom_relations":
            self.view.set_atom_relations_view(self.atom_relations_view.get_top_widget())
        elif prop.label in ("ucp_a", "ucp_b"):
            self.view.set_ucpa_view(self.ucpa_view.get_top_widget())
            self.view.set_ucpb_view(self.ucpb_view.get_top_widget())
        elif prop.label == "linked_with":
            self.reset_combo_box()
        return DummyAdapter(controller=self, prop=prop)

    def register_view(self, view):
        super(EditComponentController, self).register_view(view)

        self.layer_view = InlineObjectListStoreView(parent=view)
        self.layer_controller = EditLayerController(treemodel_property_name="layer_atoms", model=self.model, view=self.layer_view, parent=self)

        self.interlayer_view = InlineObjectListStoreView(parent=view)
        self.interlayer_controller = EditLayerController(treemodel_property_name="interlayer_atoms", model=self.model, view=self.interlayer_view, parent=self)

        self.atom_relations_view = InlineObjectListStoreView(parent=view)
        self.atom_relations_controller = EditAtomRelationsController(treemodel_property_name="atom_relations", model=self.model, view=self.atom_relations_view, parent=self)

        self.ucpa_view = EditUnitCellPropertyView(parent=view)
        self.ucpa_controller = EditUnitCellPropertyController(extra_props=[(self.model, "cell_b", "B cell length"), ], model=self.model._ucp_a, view=self.ucpa_view, parent=self)

        self.ucpb_view = EditUnitCellPropertyView(parent=view)
        self.ucpb_controller = EditUnitCellPropertyController(extra_props=[(self.model, "cell_a", "A cell length"), ], model=self.model._ucp_b, view=self.ucpb_view, parent=self)

    def register_adapters(self):
        self.update_sensitivities()

    def update_sensitivities(self):
        can_inherit = (self.model.linked_with is not None)

        def update(widget, name):
            self.view[widget].set_sensitive(not (can_inherit and getattr(self.model, "inherit_%s" % name)))
            self.view[widget].set_visible(not (can_inherit and getattr(self.model, "inherit_%s" % name)))
            self.view["component_inherit_%s" % name].set_sensitive(can_inherit)
        for name in ("default_c", "delta_c", "d001"):
            update("container_%s" % name, name)
        for name in ("interlayer_atoms", "layer_atoms", "atom_relations", "ucp_a", "ucp_b"):
            update(self.view.widget_format % name, name)


    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @Controller.observe("inherit_layer_atoms", assign=True)
    @Controller.observe("inherit_interlayer_atoms", assign=True)
    @Controller.observe("inherit_atom_relations", assign=True)
    @Controller.observe("inherit_ucp_a", assign=True)
    @Controller.observe("inherit_ucp_b", assign=True)
    @Controller.observe("inherit_d001", assign=True)
    @Controller.observe("inherit_default_c", assign=True)
    @Controller.observe("inherit_delta_c", assign=True)
    def notif_change_inherit(self, model, prop_name, info):
        self.update_sensitivities()

    @Controller.observe("linked_with", assign=True)
    def notif_linked_with_changed(self, model, prop_name, info):
        self.reset_combo_box()


    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_linked_with_changed(self, combo, user_data=None):
        itr = combo.get_active_iter()
        if itr is not None:
            val = combo.get_model().get_user_data(itr)
            self.model.linked_with = val
            self.update_sensitivities()
            return
        combo.set_active(-1)
        self.update_sensitivities()
        self.model.linked_with = None

class ComponentsController(ChildObjectListStoreController):
    """ 
        Controller for the components ObjectListStore
    """
    treemodel_property_name = "components"
    treemodel_class_type = Component
    columns = [ ("Component name", "c_name") ]
    delete_msg = "Deleting a component is irreversible!\nAre You sure you want to continue?"
    file_filters = [("Component file", get_case_insensitive_glob("*.CMP")), ]
    obj_type_map = [
        (Component, EditComponentView, EditComponentController),
    ]

    def load_components(self, filename):
        old_comps = self.get_selected_objects()
        if old_comps:
            num_oc = len(old_comps)
            new_comps = list()
            for comp in Component.load_components(filename, parent=self.model):
                comp.resolve_json_references()
                new_comps.append(comp)
            num_nc = len(new_comps)
            if num_oc != num_nc:
                DialogFactory.get_information_dialog(
                    "The number of components to import must equal the number of selected components!"
                ).run()
                return
            else:
                self.select_object(None)
                logger.info("Importing components...")
                # replace component(s):
                for old_comp, new_comp in zip(old_comps, new_comps):
                    i = self.model.components.index(old_comp)
                    self.model.components[i] = new_comp
        else:
            DialogFactory.get_information_dialog(
                "No components selected to replace!"
            ).run()

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_save_object_clicked(self, event):
        def on_accept(dialog):
            logger.info("Exporting components...")
            Component.save_components(self.get_selected_objects(), filename=dialog.filename)
        DialogFactory.get_save_dialog(
            "Export components", parent=self.view.get_toplevel(),
            filters=self.file_filters
         ).run(on_accept)
        return True

    def on_load_object_clicked(self, event):
        def on_accept(dialog):
            self.load_components(dialog.filename)
        DialogFactory.get_load_dialog(
            "Import components", parent=self.view.get_toplevel(),
            filters=self.file_filters
        ).run(on_accept)
        return True
    
    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from scipy.interpolate import interp1d

from mvc.models.properties import (
    LabeledProperty, StringProperty, StringChoiceProperty, ColorProperty,
    FloatProperty, IntegerProperty, IntegerChoiceProperty, BoolProperty,
    ListProperty, SignalProperty,
    SignalMixin, ReadOnlyMixin, SetActionMixin
)

from pyxrd.data import settings
from pyxrd.calculations.peak_detection import score_minerals
from pyxrd.generic.io import storables, Storable
from pyxrd.generic.models import ChildModel, DataModel
from pyxrd.generic.models.properties import InheritableMixin
from pyxrd.generic.models.mixins import CSVMixin
from pyxrd.generic.io.utils import unicode_open

class MineralScorer(DataModel):

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    matches_changed = SignalProperty()

    matches = ListProperty(
        default=None, text="Matches", visible=True, persistent=False,
        mix_with=(ReadOnlyMixin,)
    )

    @ListProperty(
        default=None, text="Minerals", visible=True, persistent=False,
        mix_with=(ReadOnlyMixin,)
    )
    def minerals(self):
        # Load them when accessed for the first time:
        _minerals = type(self).minerals._get(self)
        if _minerals == None:
            _minerals = list()
            with unicode_open(settings.DATA_REG.get_file_path("MINERALS")) as f:
                mineral = ""
                abbreviation = ""
                position_flag = True
                peaks = []
                for line in f:
                    line = line.replace('\n', '')
                    try:
                        number = float(line)
                        if position_flag:
                            position = number
                        else:
                            intensity = number
                            peaks.append((position, intensity))
                        position_flag = not position_flag
                    except ValueError:
                        if mineral != "":
                            _minerals.append((mineral, abbreviation, peaks))
                        position_flag = True
                        if len(line) > 25:
                            mineral = line[:24].strip()
                        if len(line) > 49:
                            abbreviation = line[49:].strip()
                        peaks = []
            sorted(_minerals, key=lambda mineral:mineral[0])
        type(self).minerals._set(self, _minerals)
        return _minerals

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, marker_peaks=[], *args, **kwargs):
        super(MineralScorer, self).__init__(*args, **kwargs)
        self._matches = []

        self.marker_peaks = marker_peaks # position, intensity

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def auto_match(self):
        self._matches = score_minerals(self.marker_peaks, self.minerals)
        self.matches_changed.emit()

    def del_match(self, index):
        if self.matches:
            del self.matches[index]
            self.matches_changed.emit()

    def add_match(self, name, abbreviation, peaks):
        matches = score_minerals(self.marker_peaks, [(name, abbreviation, peaks)])
        if len(matches):
            name, abbreviation, peaks, matches, score = matches[0]
        else:
            matches, score = [], 0.
        self.matches.append([name, abbreviation, peaks, matches, score])
        sorted(self._matches, key=lambda match: match[-1], reverse=True)
        self.matches_changed.emit()

    pass # end of class

class ThresholdSelector(ChildModel):

    # MODEL INTEL:

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    pattern = StringChoiceProperty(
        default="exp", text="Pattern",
        visible=True, persistent=False,
        choices={ "exp": "Experimental Pattern", "calc": "Calculated Pattern" },
        set_action_name="update_threshold_plot_data",
        mix_with=(SetActionMixin,)
    )

    max_threshold = FloatProperty(
        default=0.32, text="Maximum threshold",
        visible=True, persistent=False,
        minimum=0.0, maximum=1.0, widget_type="float_entry",
        set_action_name="update_threshold_plot_data",
        mix_with=(SetActionMixin,)
    )

    steps = IntegerProperty(
        default=20, text="Steps",
        visible=True, persistent=False,
        minimum=3, maximum=50,
        set_action_name="update_threshold_plot_data",
        mix_with=(SetActionMixin,)
    )

    sel_num_peaks = IntegerProperty(
        default=0, text="Selected number of peaks",
        visible=True, persistent=False,
        widget_type="label"
    )

    def set_sel_threshold(self, value):
        _sel_threshold = type(self).sel_threshold._get(self)
        if value != _sel_threshold and len(self.threshold_plot_data[0]) > 0:
            _sel_threshold = value
            if _sel_threshold >= self.threshold_plot_data[0][-1]:
                self.sel_num_peaks = self.threshold_plot_data[1][-1]
            elif _sel_threshold <= self.threshold_plot_data[0][0]:
                self.sel_num_peaks = self.threshold_plot_data[1][0]
            else:
                self.sel_num_peaks = int(interp1d(*self.threshold_plot_data)(_sel_threshold))
            type(self).sel_threshold._set(self, _sel_threshold)
    sel_threshold = FloatProperty(
        default=0.1, text="Selected threshold",
        visible=True, persistent=False,
        widget_type="float_entry",
        fset = set_sel_threshold
    )

    threshold_plot_data = LabeledProperty(
        default=None, text="Threshold plot data",
        visible=False, persistent=False
    )

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super(ThresholdSelector, self).__init__(*args, **kwargs)

        self.max_threshold = self.get_kwarg(kwargs, self.max_threshold, "max_threshold")
        self.steps = self.get_kwarg(kwargs, self.steps, "steps")
        self.sel_threshold = self.get_kwarg(kwargs, self.sel_threshold, "sel_threshold")

        if self.parent.experimental_pattern.size > 0:
            self.pattern = "exp"
        else:
            self.pattern = "calc"

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_xy(self):
        if self.pattern == "exp":
            data_x, data_y = self.parent.experimental_pattern.get_xy_data()
        elif self.pattern == "calc":
            data_x, data_y = self.parent.calculated_pattern.get_xy_data()
        if data_y.size > 0:
            data_y = data_y / np.max(data_y)
        return data_x, data_y

    _updating_plot_data = False
    def update_threshold_plot_data(self, status_dict=None):
        if self.parent is not None and not self._updating_plot_data:
            self._updating_plot_data = True
            if self.pattern == "exp":
                p, t, m = self.parent.experimental_pattern.get_best_threshold(
                            self.max_threshold, self.steps, status_dict)
            elif self.pattern == "calc":
                p, t, m = self.parent.calculated_pattern.get_best_threshold(
                            self.max_threshold, self.steps, status_dict)
            self.threshold_plot_data = p
            self.sel_threshold = t
            self.max_threshold = m
            self._updating_plot_data = False
    pass # end of class

@storables.register()
class Marker(DataModel, Storable, CSVMixin):

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "Marker"

    specimen = property(DataModel.parent.fget, DataModel.parent.fset)

    # PROPERTIES:

    #: This marker's label
    label = StringProperty(
        default="New Marker", text="Label",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether the color of this marker is inherited
    inherit_color = BoolProperty(
        default=settings.MARKER_INHERIT_COLOR, text="Inherit color",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This maker's color:
    color = ColorProperty(
        default=settings.MARKER_COLOR, text="Color",
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_color",
        inherit_from="specimen.project.display_marker_color",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Flag indicating whether the angle of this marker is inherited
    inherit_angle = BoolProperty(
        default=settings.MARKER_INHERIT_ANGLE, text="Inherit angle",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This maker's angle:
    angle = FloatProperty(
        default=settings.MARKER_ANGLE, text="Angle", widget_type="spin",
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_angle",
        inherit_from="specimen.project.display_marker_angle",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Flag indicating whether the top offset of this marker is inherited
    inherit_top_offset = BoolProperty(
        default=settings.MARKER_INHERIT_TOP_OFFSET, text="Inherit top offset",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This maker's top offset:
    top_offset = FloatProperty(
        default=settings.MARKER_TOP_OFFSET, text="Top offset", widget_type="spin",
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_top_offset",
        inherit_from="specimen.project.display_marker_top_offset",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Whether this marker is visible
    visible = BoolProperty(
        default=settings.MARKER_VISIBLE, text="Visible",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The marker's position
    position = FloatProperty(
        default=settings.MARKER_POSITION, text="Position", widget_type="spin",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The marker's x offset
    x_offset = FloatProperty(
        default=settings.MARKER_X_OFFSET, text="X offset", widget_type="spin",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: The marker's y offset
    y_offset = FloatProperty(
        default=settings.MARKER_Y_OFFSET, text="Y offset", widget_type="spin",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )

    #: Flag indicating whether the alignment of this marker is inherited
    inherit_align = BoolProperty(
        default=settings.MARKER_INHERIT_ALIGN, text="Inherit align",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This marker's alignment
    align = StringChoiceProperty(
        default=settings.MARKER_ALIGN, text="Align",
        choices=settings.MARKER_ALIGNS,
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_align",
        inherit_from="specimen.project.display_marker_align",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Flag indicating whether the base of this marker is inherited
    inherit_base = BoolProperty(
        default=settings.MARKER_INHERIT_BASE, text="Inherit base",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This marker's base
    base = IntegerChoiceProperty(
        default=settings.MARKER_BASE, text="Base",
        choices=settings.MARKER_BASES,
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_base",
        inherit_from="specimen.project.display_marker_base",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Flag indicating whether the top of this marker is inherited
    inherit_top = BoolProperty(
        default=settings.MARKER_INHERIT_TOP, text="Inherit top",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This marker's top
    top = IntegerChoiceProperty(
        default=settings.MARKER_TOP, text="Top",
        choices=settings.MARKER_TOPS,
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_top",
        inherit_from="specimen.project.display_marker_top",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    #: Flag indicating whether the line style of this marker is inherited
    inherit_style = BoolProperty(
        default=settings.MARKER_INHERIT_STYLE, text="Inherit line style",
        persistent=True, visible=True, tabular=True,
        signal_name="visuals_changed",
        mix_with=(SignalMixin,)
    )
    #: This marker's line style
    style = StringChoiceProperty(
        default=settings.MARKER_STYLE, text="Line style",
        choices=settings.MARKER_STYLES,
        persistent=True, visible=True, tabular=True, inheritable=True,
        signal_name="visuals_changed", inherit_flag="inherit_style",
        inherit_from="specimen.project.display_marker_style",
        mix_with=(InheritableMixin, SignalMixin,)
    )

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):

        my_kwargs = self.pop_kwargs(kwargs,
            "data_label", "data_visible", "data_position", "data_x_offset", "data_y_offset"
            "data_color", "data_base", "data_angle", "data_align",
            *[prop.label for prop in type(self).Meta.get_local_persistent_properties()]
        )
        super(Marker, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        self.label = self.get_kwarg(kwargs, "", "label", "data_label")
        self.visible = self.get_kwarg(kwargs, True, "visible", "data_visible")
        self.position = float(self.get_kwarg(kwargs, 0.0, "position", "data_position"))
        self.x_offset = float(self.get_kwarg(kwargs, 0.0, "x_offset", "data_x_offset"))
        self.y_offset = float(self.get_kwarg(kwargs, 0.05, "y_offset", "data_y_offset"))
        self.top_offset = float(self.get_kwarg(kwargs, 0.0, "top_offset"))
        self.color = self.get_kwarg(kwargs, settings.MARKER_COLOR, "color", "data_color")
        self.base = int(self.get_kwarg(kwargs, settings.MARKER_BASE, "base", "data_base"))
        self.angle = float(self.get_kwarg(kwargs, 0.0, "angle", "data_angle"))
        self.align = self.get_kwarg(kwargs, settings.MARKER_ALIGN, "align")
        self.style = self.get_kwarg(kwargs, settings.MARKER_STYLE, "style", "data_align")

        # if top is not set and style is not "none",
        # assume top to be "Top of plot", otherwise (style is not "none")
        # assume top to be relative to the base point (using top_offset)
        self.top = int(self.get_kwarg(kwargs, 0 if self.style == "none" else 1, "top"))

        self.inherit_align = self.get_kwarg(kwargs, True, "inherit_align")
        self.inherit_color = self.get_kwarg(kwargs, True, "inherit_color")
        self.inherit_base = self.get_kwarg(kwargs, True, "inherit_base")
        self.inherit_top = self.get_kwarg(kwargs, True, "inherit_top")
        self.inherit_top_offset = self.get_kwarg(kwargs, True, "inherit_top_offset")
        self.inherit_angle = self.get_kwarg(kwargs, True, "inherit_angle")
        self.inherit_style = self.get_kwarg(kwargs, True, "inherit_style")

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def get_nm_position(self):
        if self.parent is not None:
            return self.parent.goniometer.get_nm_from_2t(self.position)
        else:
            return 0.0

    def set_nm_position(self, position):
        if self.parent is not None:
            self.position = self.parent.goniometer.get_2t_from_nm(position)
        else:
            self.position = 0.0

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import types
import fnmatch

from .base_parser import BaseParser

class BaseGroupBarser(BaseParser):
    """
        Base class for parsers that are composed of several sub-parsers (parsers
        class attribute).
        This allows to quickly find the correct parser based on the filename.
        When a (python) file object is passed to the parse functions, the 'name'
        attribute must be present on the object, otherwise an error is raised.
        Alternatively, you can pass in a file object, together with a (fake)
        filename argument.
    """
    parsers = None

    @classmethod
    def _get_file(cls, fp, close=None):
        """
            Returns a three-tuple:
            if fp is a filename:
            filename, filename, close
            or if fp is file-like object:
            filename, file, close
            where filename can be None
            Opening of the file is left to the actual parser class if a filename
            was passed. If a file is passed, care must be taken to ensure it is
            opened in the correct mode (the __file_mode__ attribute on the 
            actual parsers class called). Therefore, it is safer to pass
            filenames to a BaseGroupParser sub-class than it is to pass a 
            file-like object. 
        """
        if isinstance(fp, str):
            return fp, fp, True if close is None else close
        else:
            return getattr(fp, 'name', None), fp, False if close is None else close

    @classmethod
    def get_parser(cls, filename, fp=None):
        if not type(filename) is str and hasattr(fp, 'name'):
            filename = fp.name
        if not type(filename) is str:
            raise TypeError("Wrong type for filename (%s), must be a string, but %s was given" % (cls.description, type(filename)))
        else:
            try:
                import gi
                gi.require_version('Gtk', '3.0')  # @UndefinedVariable
                from gi.repository import Gio  # @UnresolvedImport
                giof = Gio.File.new_for_path(filename) # @UndefinedVariable
                file_mime = giof.query_info(
                        'standard::content-type', 
                        Gio.FileQueryInfoFlags.NONE, None
                    ).get_content_type()
                del giof
            except ImportError:
                file_mime = "NONE/NONE"
                pass

            # TODO init file_mime if importerror was raised...
            for parser in cls.parsers:
                passed = False
                for mime in parser.mimetypes:
                    if file_mime.split('/')[0] == mime.split('/')[0]: # TODO if an exact match can be made, even better
                        passed = True
                        break
                    else:
                        passed = False
                for extension in parser.extensions:
                    if fnmatch.fnmatch(filename, extension):
                        passed = True
                        break
                    else:
                        passed = False
                if passed:
                    return parser
                    break # just for the sake of clarity

    @classmethod
    def parse(cls, fp, data_objects=None, close=True):
        """
            Parses the file 'fp' using one of the parsers in this group.
            'fp' should preferably be a filename (str), but can be a file-like
            object. Take care to open the file in the correct mode when passing
            a file-like object.  
        """
        filename, fp, close = cls._get_file(fp, close=close)
        parser = cls.get_parser(filename, fp=fp)
        return parser.parse(fp, data_objects=data_objects, close=close)

    @classmethod
    def setup_file_filter(cls):
        """
            Creates a file filter based on a list of extensions set in the
            'extensions' attribute of the class using the 'description' attribute
            as the name for the filter. If the 'mimetypes' attribute is also set,
            it will also set these. If additional properties are needed, this function
            should be overriden by subclasses.
        """
        try:
            import gi
            gi.require_version('Gtk', '3.0')
            from gi.repository import Gtk
        except ImportError:
            pass
        else:
            if cls.file_filter == None and cls.description and cls.parsers:
                cls.file_filter = Gtk.FileFilter()
                cls.file_filter.set_name(cls.description)
                for parser in cls.parsers:
                    for mtpe in parser.mimetypes:
                        # cls.file_filter.add_mime_type(mtpe)
                        pass
                    for expr in parser.extensions:
                        cls.file_filter.add_pattern(expr)
                setattr(cls.file_filter, "parser", cls)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
import scipy

from ..refine_method import RefineMethod
from ..refine_method_option import RefineMethodOption

import logging
logger = logging.getLogger(__name__)

MAXFUN = 15000
MAXITER = 15000
IPRINT = 0

class RefineLBFGSBRun(RefineMethod):
    """
        An implementation of the L BFGS B refinement algorithm.
    """

    name = "L BFGS B algorithm"
    description = "Refinement using the L BFGS B algorithm"
    index = 0
    disabled = False

    maxfun = RefineMethodOption('Maximum # of function calls', MAXFUN, [1, 1000000], int)
    maxiter = RefineMethodOption('Maximum # of iterations', MAXITER, [1, 1000000], int)
    iprint = RefineMethodOption('Output level [-1,0,1]', IPRINT, [-1, 1], int)

    def run(self, refiner, maxfun=MAXFUN, maxiter=MAXITER, iprint=IPRINT, **kwargs):
        """
            Refinement using the L BFGS B algorithm
        """
        solution, residual, d = scipy.optimize.fmin_l_bfgs_b(# @UnusedVariable @UndefinedVariable
            refiner.get_residual,
            refiner.history.initial_solution,
            approx_grad=True,
            bounds=refiner.ranges,
            iprint=iprint,
            epsilon=1e-4,
            callback=refiner.update,
            maxfun=maxfun, maxiter=maxiter
        )
        
        refiner.update(solution, residual=residual)
        
        logger.debug("fmin_l_bfgs_b returned: %s" % d)

    pass # end of class

class RefineBasinHoppingRun(RefineMethod):
    """
        An implementation of the Basin Hopping refinement algorithm.
    """

    name = "Basin Hopping Algorithm"
    description = "Refinement using a basin hopping algorithm"
    index = 4
    disabled = False

    niter = RefineMethodOption('Number of iterations', 100, [10, 10000], int)
    T = RefineMethodOption('Temperature criterion', 1.0, [0.0, None], int)
    stepsize = RefineMethodOption('Displacement step size', 0.5, [0.0, None], float)

    def run(self, refiner, niter=100, T=1.0, stepsize=0.5, **kwargs):
        """
            Refinement using a Basin Hopping Algorithm
        """
        vals = scipy.optimize.basinhopping(  # @UndefinedVariable
            refiner.get_residual,
            refiner.history.initial_solution,
            niter=niter,
            T=T, # this can be quite large
            stepsize=stepsize,
            minimizer_kwargs={
                'method': 'L-BFGS-B',
                'bounds': refiner.ranges,
            },
            callback=lambda s,r, a: refiner.update(s,r),
        )
        solution = np.asanyarray(vals.x)
        residual = vals.fun
        refiner.update(solution, residual=residual)

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

class MPLCanvasEventDelegator(object):
    """
        An event delegator object that supports stopping event propagation by
        returning 'True' and setting event handlers as the first.
    """

    canvas = None
    _handler_dict = None

    @classmethod
    def wrap_canvas(cls, canvas):
        delegator = getattr(canvas, "__mpl_canvas_delegator", None)
        if delegator is None:
            delegator = MPLCanvasEventDelegator(canvas)
            setattr(canvas, "__mpl_canvas_delegator", delegator)
        else:
            assert (delegator.canvas == canvas)
        return delegator

    def __init__(self, canvas):
        self.canvas = canvas
        self._handler_dict = {
            "button_press_event": [],
            "button_release_event": [],
            "draw_event": [],
            "key_press_event": [],
            "key_release_event": [],
            "motion_notify_event": [],
            "pick_event": [],
            "resize_event": [],
            "scroll_event": [],
            "figure_enter_event": [],
            "figure_leave_event": [],
            "axes_enter_event": [],
            "axes_leave_event": [],
            "close_event": [],
        }

    def _get_handlers(self, event_name):
        try:
            handlers = self._handler_dict[event_name]
        except KeyError:
            raise ValueError("Unknown event name!")
        return handlers

    def _handle_event(self, event_name):
        handlers = self._get_handlers(event_name)
        def _event_handler(event):
            for handler in handlers:
                if handler(event): # Returning true stops propagation
                    break
                else:
                    continue
        return _event_handler

    def connect(self, event_name, handler, first=False):
        """
            Connects the given handler with the given event_name.
            If first is set to True the handler will be the first in the list
            of event handlers. There is no absolute guarantee the handler will
            remain the first (e.g. later handlers can also call this with 
            first=True).
        """
        handlers = self._get_handlers(event_name)

        if len(handlers) == 0:
            setattr(self, "_event_id_%s" % event_name, self.canvas.mpl_connect(
                event_name, self._handle_event(event_name)
            ))

        if not first:
            handlers.append(handler)
        else:
            handlers.insert(0, handler)

    def disconnect(self, event_name, handler):
        """
            Disconnects the given handler for the given event_name. 
        """

        handlers = self._get_handlers(event_name)

        try:
            handlers.remove(handler)
        except ValueError:
            logger.warning(
                "Tried to disconnect an unregistered handler `%r` on `%r`" % (
                    handler, self))

        if len(handlers) == 0:
            event_id = getattr(self, "_event_id_%s" % event_name, None)
            if event_id is not None:
                self.canvas.mpl_disconnect(event_id)
            setattr(self, "_event_id_%s" % event_name, None)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Tobias Weber
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------
from .obs_wrapper import ObsWrapper

class ObsSeqWrapper (ObsWrapper):
    """
        Base class for ObsListWrapper, ObsMapWrapper, ...
        Use sub-classes, not this base class!
    """

    def __init__(self, obj, method_names):
        ObsWrapper.__init__(self, obj, method_names)

        for _m in "lt le eq ne gt ge len iter".split():
            meth = "__%s__" % _m
            assert hasattr(self._obj, meth), "Not found method %s in %s" % (meth, str(type(self._obj)))
            setattr(self.__class__, meth, getattr(self._obj, meth))
            pass
        return

    def __setitem__(self, key, val):
        self._notify_method_before(self._obj, "__setitem__", (key, val), {})
        res = self._obj.__setitem__(key, val)
        self._notify_method_after(self._obj, "__setitem__", res, (key, val), {})
        return res

    def __delitem__(self, key):
        self._notify_method_before(self._obj, "__delitem__", (key,), {})
        res = self._obj.__delitem__(key)
        self._notify_method_after(self._obj, "__delitem__", res, (key,), {})
        return res

    def __getitem__(self, key):
        return self._obj.__getitem__(key)

    pass # end of class



#!/usr/bin/env python3

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

if __name__ == "__main__":
    import logging
    logger = logging.getLogger(__name__)
    
    import os, sys
    sys.path.insert(1, os.path.join(sys.path[0], '../..'))

    from pyxrd.data.appdirs import user_log_dir

    from pyxrd.server import settings
    from pyxrd.server.pyxrd_server import PyXRDServer
    from pyxrd.server.utils import start_script
    from pyxrd.logs import setup_logging
    setup_logging(basic=True, prefix="PYRO SERVER:")

    import Pyro4
    try:
        from Pyro4.naming import NamingError
    except (AttributeError, ImportError):
        from Pyro4.errors import NamingError


    server = PyXRDServer()

    daemon = Pyro4.Daemon()

    try:
        ns = Pyro4.locateNS()
    except NamingError:
        logger.info("NamingError encountered when trying to locate the nameserver")
        log_file = os.path.join(user_log_dir('PyXRD'), 'nameserver.log')
        start_script("start_nameserver.py", auto_kill=not settings.KEEP_SERVER_ALIVE, log_file=log_file)
        ns = Pyro4.locateNS()

    server_uri = daemon.register(server)
    ns.register(settings.PYRO_NAME, server_uri) # settings.PYRO_NAME)

    try:
        daemon.requestLoop(server.loopCondition)
    finally:
        daemon.shutdown()
